import json
import os
import platform
import random
import socket
import sys
import uuid
import warnings
from collections.abc import Iterable, Mapping
from datetime import datetime, timezone
from importlib import import_module
from typing import TYPE_CHECKING, Dict, List, cast, overload

from sentry_sdk._compat import check_uwsgi_thread_support
from sentry_sdk._log_batcher import LogBatcher
from sentry_sdk._metrics_batcher import MetricsBatcher
from sentry_sdk._span_batcher import SpanBatcher
from sentry_sdk.consts import (
    DEFAULT_MAX_VALUE_LENGTH,
    DEFAULT_OPTIONS,
    INSTRUMENTER,
    SPANDATA,
    SPANSTATUS,
    VERSION,
    ClientConstructor,
)
from sentry_sdk.data_collection import (
    _map_from_send_default_pii,
    _resolve_data_collection,
)
from sentry_sdk.envelope import Envelope, Item, PayloadRef
from sentry_sdk.integrations import _DEFAULT_INTEGRATIONS, setup_integrations
from sentry_sdk.integrations.dedupe import DedupeIntegration
from sentry_sdk.monitor import Monitor
from sentry_sdk.profiler.continuous_profiler import setup_continuous_profiler
from sentry_sdk.profiler.transaction_profiler import (
    Profile,
    has_profiling_enabled,
    setup_profiler,
)
from sentry_sdk.scrubber import EventScrubber
from sentry_sdk.serializer import serialize
from sentry_sdk.sessions import SessionFlusher
from sentry_sdk.traces import SpanStatus, StreamedSpan
from sentry_sdk.tracing import trace
from sentry_sdk.tracing_utils import has_span_streaming_enabled
from sentry_sdk.transport import (
    AsyncHttpTransport,
    HttpTransportCore,
    make_transport,
)
from sentry_sdk.utils import (
    AnnotatedValue,
    ContextVar,
    capture_internal_exceptions,
    current_stacktrace,
    datetime_from_isoformat,
    env_to_bool,
    format_timestamp,
    get_before_send_log,
    get_before_send_metric,
    get_before_send_span,
    get_default_release,
    get_sdk_name,
    get_type_name,
    handle_in_app,
    has_data_collection_enabled,
    logger,
)

if TYPE_CHECKING:
    from typing import Any, Callable, Optional, Sequence, Type, TypeVar, Union

    from sentry_sdk._log_batcher import LogBatcher
    from sentry_sdk._metrics_batcher import MetricsBatcher
    from sentry_sdk._types import (
        Event,
        EventDataCategory,
        Hint,
        Log,
        Metric,
        SDKInfo,
        SerializedAttributeValue,
    )
    from sentry_sdk.integrations import Integration
    from sentry_sdk.scope import Scope
    from sentry_sdk.session import Session
    from sentry_sdk.spotlight import SpotlightClient
    from sentry_sdk.traces import StreamedSpan
    from sentry_sdk.transport import Item, Transport
    from sentry_sdk.utils import Dsn

    I = TypeVar("I", bound=Integration)  # noqa: E741

_client_init_debug = ContextVar("client_init_debug")


SDK_INFO: "SDKInfo" = {
    "name": "sentry.python",  # SDK name will be overridden after integrations have been loaded with sentry_sdk.integrations.setup_integrations()
    "version": VERSION,
    "packages": [{"name": "pypi:sentry-sdk", "version": VERSION}],
}


def _serialized_v1_attribute_to_serialized_v2_attribute(
    attribute_value: "Any",
) -> "Optional[SerializedAttributeValue]":
    if isinstance(attribute_value, bool):
        return {
            "value": attribute_value,
            "type": "boolean",
        }

    if isinstance(attribute_value, int):
        return {
            "value": attribute_value,
            "type": "integer",
        }

    if isinstance(attribute_value, float):
        return {
            "value": attribute_value,
            "type": "double",
        }

    if isinstance(attribute_value, str):
        return {
            "value": attribute_value,
            "type": "string",
        }

    if isinstance(attribute_value, list):
        if not attribute_value:
            return {"value": [], "type": "array"}

        ty = type(attribute_value[0])
        if ty in (int, str, bool, float) and all(
            type(v) is ty for v in attribute_value
        ):
            return {
                "value": attribute_value,
                "type": "array",
            }

    # Types returned when the serializer for V1 span attributes recurses into some container types.
    if isinstance(attribute_value, (dict, list)):
        return {
            "value": json.dumps(attribute_value),
            "type": "string",
        }

    return None


def _serialized_v1_span_to_serialized_v2_span(
    span: "dict[str, Any]", event: "Event"
) -> "dict[str, Any]":
    # See SpanBatcher._to_transport_format() for analogous population of all entries except "attributes".
    res: "dict[str, Any]" = {
        "status": SpanStatus.OK.value,
        "is_segment": False,
    }

    if "trace_id" in span:
        res["trace_id"] = span["trace_id"]

    if "span_id" in span:
        res["span_id"] = span["span_id"]

    if "description" in span:
        description = span["description"]

        if description is None and "op" in span:
            description = span["op"]

        res["name"] = description

    if "start_timestamp" in span:
        start_timestamp = None
        try:
            start_timestamp = datetime_from_isoformat(span["start_timestamp"])
        except Exception:
            pass

        if start_timestamp is not None:
            res["start_timestamp"] = start_timestamp.timestamp()

    if "timestamp" in span:
        end_timestamp = None
        try:
            end_timestamp = datetime_from_isoformat(span["timestamp"])
        except Exception:
            pass

        if end_timestamp is not None:
            res["end_timestamp"] = end_timestamp.timestamp()

    if "parent_span_id" in span:
        res["parent_span_id"] = span["parent_span_id"]

    if "status" in span and span["status"] != SPANSTATUS.OK:
        res["status"] = "error"

    attributes: "Dict[str, Any]" = {}

    if "op" in span:
        attributes["sentry.op"] = span["op"]
    if "origin" in span:
        attributes["sentry.origin"] = span["origin"]

    span_data = span.get("data")
    if isinstance(span_data, dict):
        attributes.update(span_data)

    span_tags = span.get("tags")
    if isinstance(span_tags, dict):
        attributes.update(span_tags)

    # See Scope._apply_user_attributes_to_telemetry() for user attributes.
    user = event.get("user")
    if isinstance(user, dict):
        if "id" in user:
            attributes["user.id"] = user["id"]
        if "username" in user:
            attributes["user.name"] = user["username"]
        if "email" in user:
            attributes["user.email"] = user["email"]

    # See Scope.set_global_attributes() for release, environment, and SDK metadata.
    if "release" in event:
        attributes["sentry.release"] = event["release"]
    if "environment" in event:
        attributes["sentry.environment"] = event["environment"]
    if "server_name" in event:
        attributes["server.address"] = event["server_name"]
    if "transaction" in event:
        attributes["sentry.segment.name"] = event["transaction"]

    trace_context = event.get("contexts", {}).get("trace", {})
    if "span_id" in trace_context:
        attributes["sentry.segment.id"] = trace_context["span_id"]

    sdk_info = event.get("sdk")
    if isinstance(sdk_info, dict):
        if "name" in sdk_info:
            attributes["sentry.sdk.name"] = sdk_info["name"]
        if "version" in sdk_info:
            attributes["sentry.sdk.version"] = sdk_info["version"]

    attributes["process.runtime.name"] = platform.python_implementation()
    attributes["process.runtime.version"] = (
        f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"
    )

    if not attributes:
        return res

    res["attributes"] = {}
    for key, value in attributes.items():
        converted_value = _serialized_v1_attribute_to_serialized_v2_attribute(value)
        if converted_value is None:
            continue

        res["attributes"][key] = converted_value

    # Remove redundant attribute, as status is stored in the status field.
    if "status" in res["attributes"]:
        del res["attributes"]["status"]

    return res


def _split_gen_ai_spans(
    event_opt: "Event",
) -> "Optional[tuple[List[Dict[str, object]], List[Dict[str, object]]]]":
    if "spans" not in event_opt:
        return None

    spans: "Any" = event_opt["spans"]
    if isinstance(spans, AnnotatedValue):
        spans = spans.value

    if not isinstance(spans, Iterable):
        return None

    non_gen_ai_spans = []
    gen_ai_spans = []
    for span in spans:
        if not isinstance(span, dict):
            non_gen_ai_spans.append(span)
            continue

        span_op = span.get("op")
        if isinstance(span_op, str) and span_op.startswith("gen_ai."):
            gen_ai_spans.append(span)
        else:
            non_gen_ai_spans.append(span)

    return non_gen_ai_spans, gen_ai_spans


def _get_options(*args: "Optional[str]", **kwargs: "Any") -> "Dict[str, Any]":
    if args and (isinstance(args[0], (bytes, str)) or args[0] is None):
        dsn: "Optional[str]" = args[0]
        args = args[1:]
    else:
        dsn = None

    if len(args) > 1:
        raise TypeError("Only single positional argument is expected")

    rv = dict(DEFAULT_OPTIONS)
    options = dict(*args, **kwargs)
    if dsn is not None and options.get("dsn") is None:
        options["dsn"] = dsn

    for key, value in options.items():
        if key not in rv:
            raise TypeError("Unknown option %r" % (key,))

        rv[key] = value

    if rv["dsn"] is None:
        rv["dsn"] = os.environ.get("SENTRY_DSN")

    if rv["release"] is None:
        rv["release"] = get_default_release()

    if rv["environment"] is None:
        rv["environment"] = os.environ.get("SENTRY_ENVIRONMENT") or "production"

    if rv["debug"] is None:
        rv["debug"] = env_to_bool(os.environ.get("SENTRY_DEBUG"), strict=True) or False

    if rv["server_name"] is None and hasattr(socket, "gethostname"):
        rv["server_name"] = socket.gethostname()

    if rv["instrumenter"] is None:
        rv["instrumenter"] = INSTRUMENTER.SENTRY

    if rv["project_root"] is None:
        try:
            project_root = os.getcwd()
        except Exception:
            project_root = None

        rv["project_root"] = project_root

    if rv["enable_tracing"] is True and rv["traces_sample_rate"] is None:
        rv["traces_sample_rate"] = 1.0

    rv["data_collection"] = _resolve_data_collection(rv)

    # Do not add the event scrubber if data collection is enabled as it can remove data that's
    # collected under data collection config
    if not has_data_collection_enabled(rv) and rv["event_scrubber"] is None:
        rv["event_scrubber"] = EventScrubber(
            send_default_pii=False
            if rv["send_default_pii"] is None
            else rv["send_default_pii"]
        )
    elif has_data_collection_enabled(rv) and rv["event_scrubber"]:
        warnings.warn(
            "Event scrubbers are not enabled when data collection configuration is provided. Ignoring event_scrubber...",
            stacklevel=2,
        )
        rv["event_scrubber"] = None

    if rv["socket_options"] and not isinstance(rv["socket_options"], list):
        logger.warning(
            "Ignoring socket_options because of unexpected format. See urllib3.HTTPConnection.socket_options for the expected format."
        )
        rv["socket_options"] = None

    if rv["keep_alive"] is None:
        rv["keep_alive"] = (
            env_to_bool(os.environ.get("SENTRY_KEEP_ALIVE"), strict=True) or False
        )

    if rv["enable_tracing"] is not None:
        warnings.warn(
            "The `enable_tracing` parameter is deprecated. Please use `traces_sample_rate` instead.",
            DeprecationWarning,
            stacklevel=2,
        )

    if rv["trace_ignore_status_codes"] and has_span_streaming_enabled(rv):
        warnings.warn(
            "The `trace_ignore_status_codes` parameter is ignored in span streaming mode.",
            stacklevel=2,
        )

    if rv["ignore_spans"] and not has_span_streaming_enabled(rv):
        warnings.warn(
            "The `ignore_spans` parameter only works when `trace_lifecycle` is set to `stream`.",
            stacklevel=2,
        )

    if rv["before_send_span"] and not has_span_streaming_enabled(rv):
        warnings.warn(
            "The `before_send_span` parameter only works when `trace_lifecycle` is set to `stream`.",
            stacklevel=2,
        )

    return rv


class BaseClient:
    """
    .. versionadded:: 2.0.0

    The basic definition of a client that is used for sending data to Sentry.
    """

    spotlight: "Optional[SpotlightClient]" = None

    def __init__(self, options: "Optional[Dict[str, Any]]" = None) -> None:
        self.options: "Dict[str, Any]" = (
            options if options is not None else DEFAULT_OPTIONS
        )

        self.transport: "Optional[Transport]" = None
        self.monitor: "Optional[Monitor]" = None
        self.log_batcher: "Optional[LogBatcher]" = None
        self.metrics_batcher: "Optional[MetricsBatcher]" = None
        self.span_batcher: "Optional[SpanBatcher]" = None
        self.integrations: "dict[str, Integration]" = {}

    def __getstate__(self, *args: "Any", **kwargs: "Any") -> "Any":
        return {"options": {}}

    def __setstate__(self, *args: "Any", **kwargs: "Any") -> None:
        pass

    @property
    def dsn(self) -> "Optional[str]":
        return None

    @property
    def parsed_dsn(self) -> "Optional[Dsn]":
        return None

    def should_send_default_pii(self) -> bool:
        return False

    def is_active(self) -> bool:
        """
        .. versionadded:: 2.0.0

        Returns whether the client is active (able to send data to Sentry)
        """
        return False

    def capture_event(self, *args: "Any", **kwargs: "Any") -> "Optional[str]":
        return None

    def _capture_log(self, log: "Log", scope: "Scope") -> None:
        pass

    def _capture_metric(self, metric: "Metric", scope: "Scope") -> None:
        pass

    def _capture_span(self, span: "StreamedSpan", scope: "Scope") -> None:
        pass

    def capture_session(self, *args: "Any", **kwargs: "Any") -> None:
        return None

    if TYPE_CHECKING:

        @overload
        def get_integration(self, name_or_class: str) -> "Optional[Integration]": ...

        @overload
        def get_integration(self, name_or_class: "type[I]") -> "Optional[I]": ...

    def get_integration(
        self, name_or_class: "Union[str, type[Integration]]"
    ) -> "Optional[Integration]":
        return None

    def close(self, *args: "Any", **kwargs: "Any") -> None:
        return None

    def flush(self, *args: "Any", **kwargs: "Any") -> None:
        return None

    async def close_async(self, *args: "Any", **kwargs: "Any") -> None:
        return None

    async def flush_async(self, *args: "Any", **kwargs: "Any") -> None:
        return None

    def __enter__(self) -> "BaseClient":
        return self

    def __exit__(self, exc_type: "Any", exc_value: "Any", tb: "Any") -> None:
        return None


class NonRecordingClient(BaseClient):
    """
    .. versionadded:: 2.0.0

    A client that does not send any events to Sentry. This is used as a fallback when the Sentry SDK is not yet initialized.
    """

    pass


class _Client(BaseClient):
    """
    The client is internally responsible for capturing the events and
    forwarding them to sentry through the configured transport.  It takes
    the client options as keyword arguments and optionally the DSN as first
    argument.

    Alias of :py:class:`sentry_sdk.Client`. (Was created for better intelisense support)
    """

    def __init__(self, *args: "Any", **kwargs: "Any") -> None:
        super(_Client, self).__init__(options=get_options(*args, **kwargs))
        self._init_impl()

    def __getstate__(self) -> "Any":
        return {"options": self.options}

    def __setstate__(self, state: "Any") -> None:
        self.options = state["options"]
        self._init_impl()

    def _setup_instrumentation(
        self, functions_to_trace: "Sequence[Dict[str, str]]"
    ) -> None:
        """
        Instruments the functions given in the list `functions_to_trace` with the `@sentry_sdk.tracing.trace` decorator.
        """
        for function in functions_to_trace:
            class_name = None
            function_qualname = function["qualified_name"]

            if "." not in function_qualname:
                logger.warning(
                    "Can not enable tracing for '%s'. Please provide the fully qualified name including the module (e.g. 'mymodule.my_function').",
                    function_qualname,
                )
                continue

            module_name, function_name = function_qualname.rsplit(".", 1)

            try:
                # Try to import module and function
                # ex: "mymodule.submodule.funcname"

                module_obj = import_module(module_name)
                function_obj = getattr(module_obj, function_name)
                setattr(module_obj, function_name, trace(function_obj))
                logger.debug("Enabled tracing for %s", function_qualname)
            except ModuleNotFoundError:
                try:
                    # Try to import a class
                    # ex: "mymodule.submodule.MyClassName.member_function"

                    module_name, class_name = module_name.rsplit(".", 1)
                    module_obj = import_module(module_name)
                    class_obj = getattr(module_obj, class_name)
                    function_obj = getattr(class_obj, function_name)
                    function_type = type(class_obj.__dict__[function_name])
                    traced_function = trace(function_obj)

                    if function_type in (staticmethod, classmethod):
                        traced_function = staticmethod(traced_function)

                    setattr(class_obj, function_name, traced_function)
                    setattr(module_obj, class_name, class_obj)
                    logger.debug("Enabled tracing for %s", function_qualname)

                except Exception as e:
                    logger.warning(
                        "Can not enable tracing for '%s'. (%s) Please check your `functions_to_trace` parameter.",
                        function_qualname,
                        e,
                    )

            except Exception as e:
                logger.warning(
                    "Can not enable tracing for '%s'. (%s) Please check your `functions_to_trace` parameter.",
                    function_qualname,
                    e,
                )

    def _init_impl(self) -> None:
        old_debug = _client_init_debug.get(False)

        def _capture_envelope(envelope: "Envelope") -> None:
            if self.spotlight is not None:
                self.spotlight.capture_envelope(envelope)
            if self.transport is not None:
                self.transport.capture_envelope(envelope)

        def _record_lost_event(
            reason: str,
            data_category: "EventDataCategory",
            item: "Optional[Item]" = None,
            quantity: int = 1,
        ) -> None:
            if self.transport is not None:
                self.transport.record_lost_event(
                    reason=reason,
                    data_category=data_category,
                    item=item,
                    quantity=quantity,
                )

        try:
            _client_init_debug.set(self.options["debug"])
            self.transport = make_transport(self.options)

            self.monitor = None
            if self.transport:
                if self.options["enable_backpressure_handling"]:
                    self.monitor = Monitor(self.transport)

            # Setup Spotlight before creating batchers so _capture_envelope can use it.
            # setup_spotlight handles all config/env var resolution per the SDK spec.
            from sentry_sdk.spotlight import setup_spotlight

            self.spotlight = setup_spotlight(self.options)
            if self.spotlight is not None and not self.options["dsn"]:
                sample_all = lambda *_args, **_kwargs: 1.0
                self.options["send_default_pii"] = True
                self.options["error_sampler"] = sample_all
                self.options["traces_sampler"] = sample_all
                self.options["profiles_sampler"] = sample_all
                # data_collection was resolved in _get_options() before this
                # spotlight override flipped send_default_pii on. Re-derive it so
                # data_collection agrees with should_send_default_pii() in
                # DSN-less spotlight mode (only when the user did not set
                # data_collection explicitly).
                if not self.options["data_collection"]["provided_by_user"]:
                    self.options["data_collection"] = _map_from_send_default_pii(
                        send_default_pii=True,
                        include_local_variables=self.options["include_local_variables"]
                        is not False,
                        include_source_context=self.options["include_source_context"]
                        is not False,
                    )

            self.session_flusher = SessionFlusher(capture_func=_capture_envelope)

            self.log_batcher = LogBatcher(
                capture_func=_capture_envelope,
                record_lost_func=_record_lost_event,
            )

            if self.options.get("enable_metrics", True) is False:
                logger.warning(
                    "The enable_metrics option has no effect and will be removed in the next major."
                )

            self.metrics_batcher = MetricsBatcher(
                capture_func=_capture_envelope,
                record_lost_func=_record_lost_event,
            )

            self.span_batcher = None
            if has_span_streaming_enabled(self.options):
                self.span_batcher = SpanBatcher(
                    capture_func=_capture_envelope,
                    record_lost_func=_record_lost_event,
                )

            max_request_body_size = ("always", "never", "small", "medium")
            if self.options["max_request_body_size"] not in max_request_body_size:
                raise ValueError(
                    "Invalid value for max_request_body_size. Must be one of {}".format(
                        max_request_body_size
                    )
                )

            if self.options["_experiments"].get("otel_powered_performance", False):
                logger.debug(
                    "[OTel] Enabling experimental OTel-powered performance monitoring."
                )
                self.options["instrumenter"] = INSTRUMENTER.OTEL
                if (
                    "sentry_sdk.integrations.opentelemetry.integration.OpenTelemetryIntegration"
                    not in _DEFAULT_INTEGRATIONS
                ):
                    _DEFAULT_INTEGRATIONS.append(
                        "sentry_sdk.integrations.opentelemetry.integration.OpenTelemetryIntegration",
                    )

            self.integrations = setup_integrations(
                self.options["integrations"],
                with_defaults=self.options["default_integrations"],
                with_auto_enabling_integrations=self.options[
                    "auto_enabling_integrations"
                ],
                disabled_integrations=self.options["disabled_integrations"],
                options=self.options,
            )

            sdk_name = get_sdk_name(list(self.integrations.keys()))
            SDK_INFO["name"] = sdk_name
            logger.debug("Setting SDK name to '%s'", sdk_name)

            if has_profiling_enabled(self.options):
                try:
                    setup_profiler(self.options)
                except Exception as e:
                    logger.debug("Can not set up profiler. (%s)", e)
            else:
                try:
                    setup_continuous_profiler(
                        self.options,
                        sdk_info=SDK_INFO,
                        capture_func=_capture_envelope,
                    )
                except Exception as e:
                    logger.debug("Can not set up continuous profiler. (%s)", e)

        finally:
            _client_init_debug.set(old_debug)

        self._setup_instrumentation(self.options.get("functions_to_trace", []))

        if (
            self.monitor
            or self.log_batcher
            or self.metrics_batcher
            or self.span_batcher
            or has_profiling_enabled(self.options)
            or isinstance(self.transport, HttpTransportCore)
        ):
            # If we have anything on that could spawn a background thread, we
            # need to check if it's safe to use them.
            check_uwsgi_thread_support()

    def is_active(self) -> bool:
        """
        .. versionadded:: 2.0.0

        Returns whether the client is active (able to send data to Sentry)
        """
        return True

    def should_send_default_pii(self) -> bool:
        """
        .. versionadded:: 2.0.0

        Returns whether the client should send default PII (Personally Identifiable Information) data to Sentry.
        """
        return self.options.get("send_default_pii") or False

    @property
    def dsn(self) -> "Optional[str]":
        """Returns the configured DSN as string."""
        return self.options["dsn"]

    @property
    def parsed_dsn(self) -> "Optional[Dsn]":
        """Returns the configured parsed DSN object."""
        return self.transport.parsed_dsn if self.transport else None

    def _prepare_event(
        self,
        event: "Event",
        hint: "Hint",
        scope: "Optional[Scope]",
    ) -> "Optional[Event]":
        previous_total_spans: "Optional[int]" = None
        previous_total_breadcrumbs: "Optional[int]" = None

        if event.get("timestamp") is None:
            event["timestamp"] = datetime.now(timezone.utc)

        is_transaction = event.get("type") == "transaction"

        if scope is not None:
            spans_before = len(cast(List[Dict[str, object]], event.get("spans", [])))
            event_ = scope.apply_to_event(event, hint, self.options)

            # one of the event/error processors returned None
            if event_ is None:
                if self.transport:
                    self.transport.record_lost_event(
                        "event_processor",
                        data_category=("transaction" if is_transaction else "error"),
                    )
                    if is_transaction:
                        self.transport.record_lost_event(
                            "event_processor",
                            data_category="span",
                            quantity=spans_before + 1,  # +1 for the transaction itself
                        )
                return None

            event = event_
            spans_delta = spans_before - len(
                cast(List[Dict[str, object]], event.get("spans", []))
            )
            span_recorder_dropped_spans: int = event.pop("_dropped_spans", 0)

            if is_transaction and self.transport is not None:
                if spans_delta > 0:
                    self.transport.record_lost_event(
                        "event_processor", data_category="span", quantity=spans_delta
                    )
                if span_recorder_dropped_spans > 0:
                    self.transport.record_lost_event(
                        "buffer_overflow",
                        data_category="span",
                        quantity=span_recorder_dropped_spans,
                    )

            dropped_spans: int = span_recorder_dropped_spans + spans_delta
            if dropped_spans > 0:
                previous_total_spans = spans_before + dropped_spans
            if scope._n_breadcrumbs_truncated > 0:
                breadcrumbs = event.get("breadcrumbs", {})
                values = (
                    breadcrumbs.get("values", [])
                    if not isinstance(breadcrumbs, AnnotatedValue)
                    else []
                )
                previous_total_breadcrumbs = (
                    len(values) + scope._n_breadcrumbs_truncated
                )

        if (
            not is_transaction
            and self.options["attach_stacktrace"]
            and "exception" not in event
            and "stacktrace" not in event
            and "threads" not in event
        ):
            with capture_internal_exceptions():
                event["threads"] = {
                    "values": [
                        {
                            "stacktrace": current_stacktrace(
                                include_local_variables=self.options.get(
                                    "include_local_variables", True
                                ),
                                max_value_length=self.options.get(
                                    "max_value_length", DEFAULT_MAX_VALUE_LENGTH
                                ),
                            ),
                            "crashed": False,
                            "current": True,
                        }
                    ]
                }

        for key in "release", "environment", "server_name", "dist":
            if event.get(key) is None and self.options[key] is not None:
                event[key] = str(self.options[key]).strip()
        if event.get("sdk") is None:
            sdk_info = dict(SDK_INFO)
            sdk_info["integrations"] = sorted(self.integrations.keys())
            event["sdk"] = sdk_info

        if event.get("platform") is None:
            event["platform"] = "python"

        event = handle_in_app(
            event,
            self.options["in_app_exclude"],
            self.options["in_app_include"],
            self.options["project_root"],
        )

        if event is not None:
            event_scrubber = self.options["event_scrubber"]
            if event_scrubber:
                event_scrubber.scrub_event(event)

        if scope is not None and scope._gen_ai_original_message_count:
            spans: "List[Dict[str, Any]] | AnnotatedValue" = event.get("spans", [])
            if isinstance(spans, list):
                for span in spans:
                    span_id = span.get("span_id", None)
                    span_data = span.get("data", {})
                    if (
                        span_id
                        and span_id in scope._gen_ai_original_message_count
                        and SPANDATA.GEN_AI_REQUEST_MESSAGES in span_data
                    ):
                        span_data[SPANDATA.GEN_AI_REQUEST_MESSAGES] = AnnotatedValue(
                            span_data[SPANDATA.GEN_AI_REQUEST_MESSAGES],
                            {"len": scope._gen_ai_original_message_count[span_id]},
                        )
        if previous_total_spans is not None:
            event["spans"] = AnnotatedValue(
                event.get("spans", []), {"len": previous_total_spans}
            )
        if previous_total_breadcrumbs is not None:
            event["breadcrumbs"] = AnnotatedValue(
                event.get("breadcrumbs", {"values": []}),
                {"len": previous_total_breadcrumbs},
            )

        # Postprocess the event here so that annotated types do
        # generally not surface in before_send
        if event is not None:
            event = cast(
                "Event",
                serialize(
                    cast("Dict[str, Any]", event),
                    max_request_body_size=self.options.get("max_request_body_size"),
                    max_value_length=self.options.get("max_value_length"),
                    custom_repr=self.options.get("custom_repr"),
                ),
            )

        before_send = self.options["before_send"]
        if (
            before_send is not None
            and event is not None
            and event.get("type") != "transaction"
        ):
            new_event = None
            with capture_internal_exceptions():
                new_event = before_send(event, hint or {})
            if new_event is None:
                logger.info("before send dropped event")
                if self.transport:
                    self.transport.record_lost_event(
                        "before_send", data_category="error"
                    )

                # If this is an exception, reset the DedupeIntegration. It still
                # remembers the dropped exception as the last exception, meaning
                # that if the same exception happens again and is not dropped
                # in before_send, it'd get dropped by DedupeIntegration.
                if event.get("exception"):
                    DedupeIntegration.reset_last_seen()

            event = new_event

        before_send_transaction = self.options["before_send_transaction"]
        if (
            before_send_transaction is not None
            and event is not None
            and event.get("type") == "transaction"
        ):
            new_event = None
            spans_before = len(cast(List[Dict[str, object]], event.get("spans", [])))
            with capture_internal_exceptions():
                new_event = before_send_transaction(event, hint or {})
            if new_event is None:
                logger.info("before send transaction dropped event")
                if self.transport:
                    self.transport.record_lost_event(
                        reason="before_send", data_category="transaction"
                    )
                    self.transport.record_lost_event(
                        reason="before_send",
                        data_category="span",
                        quantity=spans_before + 1,  # +1 for the transaction itself
                    )
            else:
                spans_delta = spans_before - len(new_event.get("spans", []))
                if spans_delta > 0 and self.transport is not None:
                    self.transport.record_lost_event(
                        reason="before_send", data_category="span", quantity=spans_delta
                    )

            event = new_event

        return event

    def _is_ignored_error(self, event: "Event", hint: "Hint") -> bool:
        exc_info = hint.get("exc_info")
        if exc_info is None:
            return False

        error = exc_info[0]
        error_type_name = get_type_name(exc_info[0])
        error_full_name = "%s.%s" % (exc_info[0].__module__, error_type_name)

        for ignored_error in self.options["ignore_errors"]:
            # String types are matched against the type name in the
            # exception only
            if isinstance(ignored_error, str):
                if ignored_error == error_full_name or ignored_error == error_type_name:
                    return True
            else:
                if issubclass(error, ignored_error):
                    return True

        return False

    def _should_capture(
        self,
        event: "Event",
        hint: "Hint",
        scope: "Optional[Scope]" = None,
    ) -> bool:
        # Transactions are sampled independent of error events.
        is_transaction = event.get("type") == "transaction"
        if is_transaction:
            return True

        ignoring_prevents_recursion = scope is not None and not scope._should_capture
        if ignoring_prevents_recursion:
            return False

        ignored_by_config_option = self._is_ignored_error(event, hint)
        if ignored_by_config_option:
            return False

        return True

    def _should_sample_error(
        self,
        event: "Event",
        hint: "Hint",
    ) -> bool:
        error_sampler = self.options.get("error_sampler", None)

        if callable(error_sampler):
            with capture_internal_exceptions():
                sample_rate = error_sampler(event, hint)
        else:
            sample_rate = self.options["sample_rate"]

        try:
            not_in_sample_rate = sample_rate < 1.0 and random.random() >= sample_rate
        except NameError:
            logger.warning(
                "The provided error_sampler raised an error. Defaulting to sampling the event."
            )

            # If the error_sampler raised an error, we should sample the event, since the default behavior
            # (when no sample_rate or error_sampler is provided) is to sample all events.
            not_in_sample_rate = False
        except TypeError:
            parameter, verb = (
                ("error_sampler", "returned")
                if callable(error_sampler)
                else ("sample_rate", "contains")
            )
            logger.warning(
                "The provided %s %s an invalid value of %s. The value should be a float or a bool. Defaulting to sampling the event."
                % (parameter, verb, repr(sample_rate))
            )

            # If the sample_rate has an invalid value, we should sample the event, since the default behavior
            # (when no sample_rate or error_sampler is provided) is to sample all events.
            not_in_sample_rate = False

        if not_in_sample_rate:
            # because we will not sample this event, record a "lost event".
            if self.transport:
                self.transport.record_lost_event("sample_rate", data_category="error")

            return False

        return True

    def _update_session_from_event(
        self,
        session: "Session",
        event: "Event",
    ) -> None:
        crashed = False
        errored = False
        user_agent = None

        exceptions = (event.get("exception") or {}).get("values")
        if exceptions:
            errored = True
            for error in exceptions:
                if isinstance(error, AnnotatedValue):
                    error = error.value or {}
                mechanism = error.get("mechanism")
                if isinstance(mechanism, Mapping) and mechanism.get("handled") is False:
                    crashed = True
                    break

        user = event.get("user")

        if session.user_agent is None:
            headers = (event.get("request") or {}).get("headers")
            headers_dict = headers if isinstance(headers, dict) else {}
            for k, v in headers_dict.items():
                if k.lower() == "user-agent":
                    user_agent = v
                    break

        session.update(
            status="crashed" if crashed else None,
            user=user,
            user_agent=user_agent,
            errors=session.errors + (errored or crashed),
        )

    def capture_event(
        self,
        event: "Event",
        hint: "Optional[Hint]" = None,
        scope: "Optional[Scope]" = None,
    ) -> "Optional[str]":
        """Captures an event.

        :param event: A ready-made event that can be directly sent to Sentry.

        :param hint: Contains metadata about the event that can be read from `before_send`, such as the original exception object or a HTTP request object.

        :param scope: An optional :py:class:`sentry_sdk.Scope` to apply to events.

        :returns: An event ID. May be `None` if there is no DSN set or of if the SDK decided to discard the event for other reasons. In such situations setting `debug=True` on `init()` may help.
        """
        hint = dict(hint or ())

        if not self._should_capture(event, hint, scope):
            return None

        profile = event.pop("profile", None)

        event_id = event.get("event_id")
        if event_id is None:
            event["event_id"] = event_id = uuid.uuid4().hex

        span_recorder_has_gen_ai_span = event.pop("_has_gen_ai_span", False)
        event_opt = self._prepare_event(event, hint, scope)
        if event_opt is None:
            return None

        # whenever we capture an event we also check if the session needs
        # to be updated based on that information.
        session = scope._session if scope else None
        if session:
            self._update_session_from_event(session, event)

        is_transaction = event_opt.get("type") == "transaction"
        is_checkin = event_opt.get("type") == "check_in"

        if (
            not is_transaction
            and not is_checkin
            and not self._should_sample_error(event, hint)
        ):
            return None

        attachments = hint.get("attachments")

        trace_context = event_opt.get("contexts", {}).get("trace") or {}
        dynamic_sampling_context = trace_context.pop("dynamic_sampling_context", {})

        headers: "dict[str, object]" = {
            "event_id": event_opt["event_id"],
            "sent_at": format_timestamp(datetime.now(timezone.utc)),
        }

        if dynamic_sampling_context:
            headers["trace"] = dynamic_sampling_context

        envelope = Envelope(headers=headers)

        if is_transaction and isinstance(profile, Profile):
            envelope.add_profile(profile.to_json(event_opt, self.options))

        if is_transaction and not span_recorder_has_gen_ai_span:
            envelope.add_transaction(event_opt)
        elif is_transaction:
            split_spans = _split_gen_ai_spans(event_opt)
            if split_spans is None or not split_spans[1]:
                envelope.add_transaction(event_opt)
            else:
                non_gen_ai_spans, gen_ai_spans = split_spans

                event_opt["spans"] = non_gen_ai_spans
                envelope.add_transaction(event_opt)

                converted_gen_ai_spans = [
                    _serialized_v1_span_to_serialized_v2_span(span, event_opt)
                    for span in gen_ai_spans
                    if isinstance(span, dict)
                ]

                envelope.add_item(
                    Item(
                        type=SpanBatcher.TYPE,
                        content_type=SpanBatcher.CONTENT_TYPE,
                        headers={
                            "item_count": len(converted_gen_ai_spans),
                        },
                        payload=PayloadRef(
                            json={
                                "version": 2,
                                "items": converted_gen_ai_spans,
                            },
                        ),
                    )
                )

        elif is_checkin:
            envelope.add_checkin(event_opt)
        else:
            envelope.add_event(event_opt)

        for attachment in attachments or ():
            envelope.add_item(attachment.to_envelope_item())

        return_value = None
        if self.spotlight:
            self.spotlight.capture_envelope(envelope)
            return_value = event_id

        if self.transport is not None:
            self.transport.capture_envelope(envelope)
            return_value = event_id

        return return_value

    def _capture_telemetry(
        self,
        telemetry: "Optional[Union[Log, Metric, StreamedSpan]]",
        ty: str,
        scope: "Scope",
    ) -> None:
        """
        Capture attributes-based telemetry (logs, metrics, streamed spans).

        Apply any attributes set on the scope to it, and run the user's
        before_send_{telemetry} on it, if applicable.
        """
        if telemetry is None:
            return

        scope.apply_to_telemetry(telemetry)

        before_send = None

        if ty == "log":
            before_send = get_before_send_log(self.options)
            serialized = telemetry

        elif ty == "metric":
            before_send = get_before_send_metric(self.options)
            serialized = telemetry

        elif ty == "span":
            before_send = get_before_send_span(self.options)
            serialized = telemetry._to_json()  # type: ignore[union-attr]

        if before_send is not None:
            exception_raised_in_before_send_func = False
            with capture_internal_exceptions():
                try:
                    serialized = before_send(serialized, {})  # type: ignore[arg-type]
                except Exception:
                    exception_raised_in_before_send_func = True
                    raise

            if ty in ("log", "metric"):
                # We are ok with dropping metrics and logs when an exception is raised
                # because we allow users to drop them in their respect before_send_*
                # functions.
                if exception_raised_in_before_send_func:
                    return
                # Logs and metrics can be dropped in their respective
                # before_send, so if we get None, don't queue them for sending.
                if serialized is None:
                    return

            elif ty == "span" and isinstance(telemetry, StreamedSpan):
                # Reset the span to its original value before we attempted
                # to call the `before_send_span` callback
                if exception_raised_in_before_send_func:
                    serialized = telemetry._to_json()

                # Spans can't be dropped in before_send_span by design. They can
                # be altered though (e.g. to sanitize). Only allow changes to
                # name and attributes.
                if isinstance(serialized, dict) and "name" in serialized:
                    telemetry.name = serialized["name"]  # type: ignore[typeddict-item]
                    telemetry._attributes = {}
                    for k, v in (serialized.get("attributes") or {}).items():
                        telemetry.set_attribute(k, v)

                else:
                    logger.debug(
                        "[Tracing] Invalid return value from before_send_span. Keeping original span."
                    )

                serialized = telemetry._to_json()

        batcher = None
        if ty == "log":
            batcher = self.log_batcher

        elif ty == "metric":
            batcher = self.metrics_batcher

        elif ty == "span":
            # We need a reference to the segment span in the batcher to populate
            # the dynamic sampling context (DSC)
            serialized["_segment_span"] = telemetry._segment  # type: ignore
            batcher = self.span_batcher

        if batcher is not None:
            batcher.add(serialized)  # type: ignore

    def _capture_log(self, log: "Optional[Log]", scope: "Scope") -> None:
        self._capture_telemetry(log, "log", scope)

    def _capture_metric(self, metric: "Optional[Metric]", scope: "Scope") -> None:
        self._capture_telemetry(metric, "metric", scope)

    def _capture_span(self, span: "Optional[StreamedSpan]", scope: "Scope") -> None:
        self._capture_telemetry(span, "span", scope)

    def capture_session(
        self,
        session: "Session",
    ) -> None:
        if not session.release:
            logger.info("Discarded session update because of missing release")
        else:
            self.session_flusher.add_session(session)

    if TYPE_CHECKING:

        @overload
        def get_integration(self, name_or_class: str) -> "Optional[Integration]": ...

        @overload
        def get_integration(self, name_or_class: "type[I]") -> "Optional[I]": ...

    def get_integration(
        self,
        name_or_class: "Union[str, Type[Integration]]",
    ) -> "Optional[Integration]":
        """Returns the integration for this client by name or class.
        If the client does not have that integration then `None` is returned.
        """
        if isinstance(name_or_class, str):
            integration_name = name_or_class
        elif name_or_class.identifier is not None:
            integration_name = name_or_class.identifier
        else:
            raise ValueError("Integration has no name")

        return self.integrations.get(integration_name)

    def _has_async_transport(self) -> bool:
        """Check if the current transport is async."""
        return isinstance(self.transport, AsyncHttpTransport)

    @property
    def _batchers(self) -> "tuple[Any, ...]":
        return tuple(
            b
            for b in (self.log_batcher, self.metrics_batcher, self.span_batcher)
            if b is not None
        )

    def _close_components(self) -> None:
        """Kill all client components in the correct order."""
        self.session_flusher.kill()
        for b in self._batchers:
            b.kill()
        if self.monitor:
            self.monitor.kill()

    def _flush_components(self) -> None:
        """Flush all client components."""
        self.session_flusher.flush()
        for b in self._batchers:
            b.flush()

    def close(
        self,
        timeout: "Optional[float]" = None,
        callback: "Optional[Callable[[int, float], None]]" = None,
    ) -> None:
        """
        Close the client and shut down the transport. Arguments have the same
        semantics as :py:meth:`Client.flush`.
        """
        if self.transport is not None:
            if self._has_async_transport():
                warnings.warn(
                    "close() used with AsyncHttpTransport. Use close_async() instead.",
                    stacklevel=2,
                )
                self._flush_components()
            else:
                self.flush(timeout=timeout, callback=callback)
            self._close_components()
            self.transport.kill()
            self.transport = None

    async def close_async(
        self,
        timeout: "Optional[float]" = None,
        callback: "Optional[Callable[[int, float], None]]" = None,
    ) -> None:
        """
        Asynchronously close the client and shut down the transport. Arguments have the same
        semantics as :py:meth:`Client.flush_async`.
        """
        if self.transport is not None:
            if not self._has_async_transport():
                logger.debug(
                    "close_async() used with non-async transport, aborting. Please use close() instead."
                )
                return
            await self.flush_async(timeout=timeout, callback=callback)
            self._close_components()
            kill_task = self.transport.kill()  # type: ignore
            if kill_task is not None:
                await kill_task
            self.transport = None

    def flush(
        self,
        timeout: "Optional[float]" = None,
        callback: "Optional[Callable[[int, float], None]]" = None,
    ) -> None:
        """
        Wait for the current events to be sent.

        :param timeout: Wait for at most `timeout` seconds. If no `timeout` is provided, the `shutdown_timeout` option value is used.

        :param callback: Is invoked with the number of pending events and the configured timeout.
        """
        if self.transport is not None:
            if self._has_async_transport():
                warnings.warn(
                    "flush() used with AsyncHttpTransport. Use flush_async() instead.",
                    stacklevel=2,
                )
                return
            if timeout is None:
                timeout = self.options["shutdown_timeout"]
            self._flush_components()

            self.transport.flush(timeout=timeout, callback=callback)

    async def flush_async(
        self,
        timeout: "Optional[float]" = None,
        callback: "Optional[Callable[[int, float], None]]" = None,
    ) -> None:
        """
        Asynchronously wait for the current events to be sent.

        :param timeout: Wait for at most `timeout` seconds. If no `timeout` is provided, the `shutdown_timeout` option value is used.

        :param callback: Is invoked with the number of pending events and the configured timeout.
        """
        if self.transport is not None:
            if not self._has_async_transport():
                logger.debug(
                    "flush_async() used with non-async transport, aborting. Please use flush() instead."
                )
                return
            if timeout is None:
                timeout = self.options["shutdown_timeout"]
            self._flush_components()
            flush_task = self.transport.flush(timeout=timeout, callback=callback)  # type: ignore
            if flush_task is not None:
                await flush_task

    def __enter__(self) -> "_Client":
        return self

    def __exit__(self, exc_type: "Any", exc_value: "Any", tb: "Any") -> None:
        self.close()

    async def __aenter__(self) -> "_Client":
        return self

    async def __aexit__(self, exc_type: "Any", exc_value: "Any", tb: "Any") -> None:
        await self.close_async()


from typing import TYPE_CHECKING

if TYPE_CHECKING:
    # Make mypy, PyCharm and other static analyzers think `get_options` is a
    # type to have nicer autocompletion for params.
    #
    # Use `ClientConstructor` to define the argument types of `init` and
    # `Dict[str, Any]` to tell static analyzers about the return type.

    class get_options(ClientConstructor, Dict[str, Any]):  # noqa: N801
        pass

    class Client(ClientConstructor, _Client):
        pass

else:
    # Alias `get_options` for actual usage. Go through the lambda indirection
    # to throw PyCharm off of the weakly typed signature (it would otherwise
    # discover both the weakly typed signature of `_init` and our faked `init`
    # type).

    get_options = (lambda: _get_options)()
    Client = (lambda: _Client)()



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any


_SENTINEL = object()


class LRUCache:
    def __init__(self, max_size: int) -> None:
        if max_size <= 0:
            raise AssertionError(f"invalid max_size: {max_size}")
        self.max_size = max_size
        self._data: "dict[Any, Any]" = {}
        self.hits = self.misses = 0
        self.full = False

    def set(self, key: "Any", value: "Any") -> None:
        current = self._data.pop(key, _SENTINEL)
        if current is not _SENTINEL:
            self._data[key] = value
        elif self.full:
            self._data.pop(next(iter(self._data)))
            self._data[key] = value
        else:
            self._data[key] = value
        self.full = len(self._data) >= self.max_size

    def get(self, key: "Any", default: "Any" = None) -> "Any":
        try:
            ret = self._data.pop(key)
        except KeyError:
            self.misses += 1
            ret = default
        else:
            self.hits += 1
            self._data[key] = ret

        return ret

    def get_all(self) -> "list[tuple[Any, Any]]":
        return list(self._data.items())



import json
from typing import TYPE_CHECKING

import sentry_sdk
from sentry_sdk.consts import OP, SPANDATA
from sentry_sdk.integrations import DidNotEnable
from sentry_sdk.traces import StreamedSpan
from sentry_sdk.tracing_utils import has_span_streaming_enabled
from sentry_sdk.utils import has_data_collection_enabled

try:
    from agents import (
        CodeInterpreterTool,
        ComputerTool,
        FileSearchTool,
        FunctionTool,
        HostedMCPTool,
        ImageGenerationTool,
        LocalShellTool,
        WebSearchTool,
    )
except ImportError:
    raise DidNotEnable("OpenAI Agents not installed")

try:
    from agents import ApplyPatchTool, ShellTool
except ImportError:
    ShellTool = None
    ApplyPatchTool = None

try:
    from agents import ToolSearchTool
except ImportError:
    ToolSearchTool = None

try:
    from agents import CustomTool
except ImportError:
    CustomTool = None

try:
    from agents import ProgrammaticToolCallingTool
except ImportError:
    ProgrammaticToolCallingTool = None


from ..consts import SPAN_ORIGIN
from ..utils import (
    _set_agent_data,
    _set_input_data,
    _set_output_data,
    _set_usage_data,
)

if TYPE_CHECKING:
    from typing import Any, Optional, Union

    from agents import Agent, Tool

    from sentry_sdk._types import ToolDefinition


def _transform_tool_definitions(tools: "list[Tool]") -> "list[ToolDefinition]":
    """
    Transform tool definitions to the schema used by the "gen_ai.tool.definitions" attribute.
    Includes special handling for tools where the type includes a name, description or parameters.

    Note: These are generated from the same OpenAPI spec as Responses API tools.
    """
    if not isinstance(tools, list):
        return []

    tool_definitions: "list[ToolDefinition]" = []
    for tool in tools:
        if isinstance(tool, FunctionTool):
            tool_definitions.append(
                {
                    "type": "function",
                    "name": tool.name,
                    "description": tool.description,
                    "parameters": tool.params_json_schema,
                }
            )
            continue

        if isinstance(tool, WebSearchTool):
            tool_definitions.append(
                {
                    "type": "web_search",
                    "name": tool.name,
                }
            )
            continue

        if isinstance(tool, FileSearchTool):
            tool_definitions.append(
                {
                    "type": "file_search",
                    "name": tool.name,
                }
            )
            continue

        if isinstance(tool, ComputerTool):
            tool_definitions.append(
                {
                    "type": "computer",
                    "name": tool.name,
                }
            )
            continue

        if CustomTool is not None and isinstance(tool, CustomTool):
            tool_definitions.append(
                {
                    "type": "custom",
                    "name": tool.name,
                    "description": tool.description,
                }
            )
            continue

        if isinstance(tool, HostedMCPTool):
            tool_definitions.append(
                {
                    "type": "mcp",
                    "name": tool.name,
                }
            )
            continue

        if ApplyPatchTool is not None and isinstance(tool, ApplyPatchTool):
            tool_definitions.append(
                {
                    "type": "apply_patch",
                    "name": tool.name,
                }
            )
            continue

        if ShellTool is not None and isinstance(tool, ShellTool):
            tool_definitions.append(
                {
                    "type": "shell",
                    "name": tool.name,
                }
            )
            continue

        if isinstance(tool, ImageGenerationTool):
            tool_definitions.append(
                {
                    "type": "image_generation",
                    "name": tool.name,
                }
            )
            continue

        if isinstance(tool, CodeInterpreterTool):
            tool_definitions.append(
                {
                    "type": "code_interpreter",
                    "name": tool.name,
                }
            )
            continue

        if isinstance(tool, LocalShellTool):
            tool_definitions.append(
                {
                    "type": "local_shell",
                    "name": tool.name,
                }
            )
            continue

        if ToolSearchTool is not None and isinstance(tool, ToolSearchTool):
            tool_definitions.append(
                {
                    "type": "tool_search",
                    "name": tool.name,
                }
            )
            continue

        if ProgrammaticToolCallingTool is not None and isinstance(
            tool, ProgrammaticToolCallingTool
        ):
            tool_definitions.append(
                {
                    "type": "programmatic_tool_calling",
                    "name": tool.name,
                }
            )
            continue

    return tool_definitions


def ai_client_span(
    agent: "Agent", get_response_kwargs: "dict[str, Any]"
) -> "Union[sentry_sdk.tracing.Span, StreamedSpan]":
    # TODO-anton: implement other types of operations. Now "chat" is hardcoded.
    # Get model name from agent.model or fall back to request model (for when agent.model is None/default)
    model_name = None
    if agent.model:
        model_name = agent.model.model if hasattr(agent.model, "model") else agent.model
    elif hasattr(agent, "_sentry_request_model"):
        model_name = agent._sentry_request_model

    client_options = sentry_sdk.get_client().options

    span_streaming = has_span_streaming_enabled(client_options)
    if span_streaming:
        span = sentry_sdk.traces.start_span(
            name=f"chat {model_name}",
            attributes={
                "sentry.op": OP.GEN_AI_CHAT,
                "sentry.origin": SPAN_ORIGIN,
                SPANDATA.GEN_AI_OPERATION_NAME: "chat",
            },
        )

        set_on_span = span.set_attribute
    else:
        span = sentry_sdk.start_span(
            op=OP.GEN_AI_CHAT,
            name=f"chat {model_name}",
            origin=SPAN_ORIGIN,
        )
        # TODO-anton: remove hardcoded stuff and replace something that also works for embedding and so on
        span.set_data(SPANDATA.GEN_AI_OPERATION_NAME, "chat")

        set_on_span = span.set_data

    _set_agent_data(span, agent)

    if has_data_collection_enabled(client_options):
        if client_options["data_collection"]["gen_ai"]["inputs"]:
            set_on_span(
                SPANDATA.GEN_AI_TOOL_DEFINITIONS,
                json.dumps(_transform_tool_definitions(agent.tools)),
            )
    else:
        # This is set unconditionally prior to data collection being introduced.
        # Remove this block once data collection is fully rolled out
        set_on_span(
            SPANDATA.GEN_AI_TOOL_DEFINITIONS,
            json.dumps(_transform_tool_definitions(agent.tools)),
        )

    _set_input_data(span, get_response_kwargs)

    return span


def update_ai_client_span(
    span: "Union[sentry_sdk.tracing.Span, StreamedSpan]",
    response: "Any",
    response_model: "Optional[str]" = None,
    agent: "Optional[Agent]" = None,
) -> None:
    """Update AI client span with response data (works for streaming and non-streaming)."""
    if hasattr(response, "usage") and response.usage:
        _set_usage_data(span, response.usage)

    if hasattr(response, "output") and response.output:
        _set_output_data(span, response)

    set_on_span = (
        span.set_attribute if isinstance(span, StreamedSpan) else span.set_data
    )

    if response_model is not None:
        set_on_span(SPANDATA.GEN_AI_RESPONSE_MODEL, response_model)
    elif hasattr(response, "model") and response.model:
        set_on_span(SPANDATA.GEN_AI_RESPONSE_MODEL, str(response.model))

    # Set conversation ID from agent if available
    if agent:
        conv_id = getattr(agent, "_sentry_conversation_id", None)
        if conv_id:
            set_on_span(SPANDATA.GEN_AI_CONVERSATION_ID, conv_id)



from typing import TYPE_CHECKING

import sentry_sdk
from sentry_sdk.consts import OP
from sentry_sdk.integrations import DidNotEnable
from sentry_sdk.integrations.grpc.consts import SPAN_ORIGIN
from sentry_sdk.traces import SegmentNameSource
from sentry_sdk.tracing import TransactionSource
from sentry_sdk.tracing_utils import has_span_streaming_enabled

if TYPE_CHECKING:
    from typing import Callable, Optional

    from google.protobuf.message import Message

try:
    import grpc
    from grpc import HandlerCallDetails, RpcMethodHandler, ServicerContext
except ImportError:
    raise DidNotEnable("grpcio is not installed")


class ServerInterceptor(grpc.ServerInterceptor):  # type: ignore
    def __init__(
        self: "ServerInterceptor",
        find_name: "Optional[Callable[[ServicerContext], str]]" = None,
    ) -> None:
        self._custom_find_name = find_name

        super().__init__()

    def intercept_service(
        self: "ServerInterceptor",
        continuation: "Callable[[HandlerCallDetails], RpcMethodHandler]",
        handler_call_details: "HandlerCallDetails",
    ) -> "RpcMethodHandler":
        handler = continuation(handler_call_details)
        if not handler or not handler.unary_unary:
            return handler

        method_name = handler_call_details.method
        custom_find_name = self._custom_find_name

        def behavior(request: "Message", context: "ServicerContext") -> "Message":
            with sentry_sdk.isolation_scope():
                name = custom_find_name(context) if custom_find_name else method_name

                if name:
                    metadata = dict(context.invocation_metadata())

                    span_streaming = has_span_streaming_enabled(
                        sentry_sdk.get_client().options
                    )
                    if span_streaming:
                        sentry_sdk.traces.continue_trace(metadata)

                        with sentry_sdk.traces.start_span(
                            name=name,
                            attributes={
                                "sentry.op": OP.GRPC_SERVER,
                                "sentry.segment.name.source": SegmentNameSource.CUSTOM.value,
                                "sentry.origin": SPAN_ORIGIN,
                            },
                            parent_span=None,
                        ):
                            try:
                                return handler.unary_unary(request, context)
                            except BaseException as e:
                                raise e
                    else:
                        transaction = sentry_sdk.continue_trace(
                            metadata,
                            op=OP.GRPC_SERVER,
                            name=name,
                            source=TransactionSource.CUSTOM,
                            origin=SPAN_ORIGIN,
                        )

                        with sentry_sdk.start_transaction(transaction=transaction):
                            try:
                                return handler.unary_unary(request, context)
                            except BaseException as e:
                                raise e
                else:
                    return handler.unary_unary(request, context)

        return grpc.unary_unary_rpc_method_handler(
            behavior,
            request_deserializer=handler.request_deserializer,
            response_serializer=handler.response_serializer,
        )



import functools
from typing import TYPE_CHECKING

import sentry_sdk
from sentry_sdk.consts import OP
from sentry_sdk.traces import StreamedSpan
from sentry_sdk.tracing_utils import has_span_streaming_enabled

if TYPE_CHECKING:
    from typing import Any

from sentry_sdk.integrations.django.asgi import iscoroutinefunction, wrap_async_view


def patch_views() -> None:
    from django.core.handlers.base import BaseHandler
    from django.template.response import SimpleTemplateResponse

    from sentry_sdk.integrations.django import DjangoIntegration

    old_make_view_atomic = BaseHandler.make_view_atomic
    old_render = SimpleTemplateResponse.render

    def sentry_patched_render(self: "SimpleTemplateResponse") -> "Any":
        span_streaming = has_span_streaming_enabled(sentry_sdk.get_client().options)
        if span_streaming:
            if sentry_sdk.traces.get_current_span() is None:
                return old_render(self)
            with sentry_sdk.traces.start_span(
                name="serialize response",
                attributes={
                    "sentry.op": OP.VIEW_RESPONSE_RENDER,
                    "sentry.origin": DjangoIntegration.origin,
                },
            ):
                return old_render(self)
        else:
            with sentry_sdk.start_span(
                op=OP.VIEW_RESPONSE_RENDER,
                name="serialize response",
                origin=DjangoIntegration.origin,
            ):
                return old_render(self)

    @functools.wraps(old_make_view_atomic)
    def sentry_patched_make_view_atomic(
        self: "Any", *args: "Any", **kwargs: "Any"
    ) -> "Any":
        callback = old_make_view_atomic(self, *args, **kwargs)

        # XXX: The wrapper function is created for every request. Find more
        # efficient way to wrap views (or build a cache?)

        integration = sentry_sdk.get_client().get_integration(DjangoIntegration)
        if integration is None:
            return callback

        if iscoroutinefunction(callback):
            return wrap_async_view(callback)

        return _wrap_sync_view(callback)

    SimpleTemplateResponse.render = sentry_patched_render
    BaseHandler.make_view_atomic = sentry_patched_make_view_atomic


def _wrap_sync_view(callback: "Any") -> "Any":
    from sentry_sdk.integrations.django import DjangoIntegration

    @functools.wraps(callback)
    def sentry_wrapped_callback(request: "Any", *args: "Any", **kwargs: "Any") -> "Any":
        client = sentry_sdk.get_client()
        span_streaming = has_span_streaming_enabled(client.options)
        current_scope = sentry_sdk.get_current_scope()
        if span_streaming:
            current_span = current_scope.streamed_span
            if type(current_span) is StreamedSpan:
                segment = current_span._segment
                segment._update_active_thread()
        else:
            if current_scope.transaction is not None:
                current_scope.transaction.update_active_thread()

        sentry_scope = sentry_sdk.get_isolation_scope()
        # set the active thread id to the handler thread for sync views
        # this isn't necessary for async views since that runs on main
        if sentry_scope.profile is not None:
            sentry_scope.profile.update_active_thread_id()

        integration = client.get_integration(DjangoIntegration)
        if not integration or not integration.middleware_spans:
            return callback(request, *args, **kwargs)

        if span_streaming:
            if sentry_sdk.traces.get_current_span() is None:
                return callback(request, *args, **kwargs)
            with sentry_sdk.traces.start_span(
                name=request.resolver_match.view_name,
                attributes={
                    "sentry.op": OP.VIEW_RENDER,
                    "sentry.origin": DjangoIntegration.origin,
                },
            ):
                return callback(request, *args, **kwargs)
        else:
            with sentry_sdk.start_span(
                op=OP.VIEW_RENDER,
                name=request.resolver_match.view_name,
                origin=DjangoIntegration.origin,
            ):
                return callback(request, *args, **kwargs)

    return sentry_wrapped_callback



import functools
import sys
from copy import deepcopy
from datetime import datetime, timedelta, timezone
from os import environ
from typing import TYPE_CHECKING

import sentry_sdk
from sentry_sdk.api import continue_trace
from sentry_sdk.consts import OP
from sentry_sdk.data_collection import _apply_data_collection_filtering_to_query_string
from sentry_sdk.integrations import Integration
from sentry_sdk.integrations._wsgi_common import _filter_headers
from sentry_sdk.integrations.cloud_resource_context import CLOUD_PROVIDER
from sentry_sdk.scope import Scope, should_send_default_pii
from sentry_sdk.traces import SegmentNameSource
from sentry_sdk.tracing import TransactionSource
from sentry_sdk.tracing_utils import has_span_streaming_enabled
from sentry_sdk.utils import (
    AnnotatedValue,
    TimeoutThread,
    capture_internal_exceptions,
    event_from_exception,
    has_data_collection_enabled,
    logger,
    reraise,
)

# Constants
TIMEOUT_WARNING_BUFFER = 1.5  # Buffer time required to send timeout warning to Sentry
MILLIS_TO_SECONDS = 1000.0

if TYPE_CHECKING:
    from typing import Any, Callable, Optional, TypeVar

    from sentry_sdk._types import Event, EventProcessor, Hint

    F = TypeVar("F", bound=Callable[..., Any])


def _wrap_func(func: "F") -> "F":
    @functools.wraps(func)
    def sentry_func(
        functionhandler: "Any", gcp_event: "Any", *args: "Any", **kwargs: "Any"
    ) -> "Any":
        client = sentry_sdk.get_client()

        integration = client.get_integration(GcpIntegration)
        if integration is None:
            return func(functionhandler, gcp_event, *args, **kwargs)

        configured_time = environ.get("FUNCTION_TIMEOUT_SEC")
        if not configured_time:
            logger.debug(
                "The configured timeout could not be fetched from Cloud Functions configuration."
            )
            return func(functionhandler, gcp_event, *args, **kwargs)

        configured_time = int(configured_time)

        initial_time = datetime.now(timezone.utc)

        with sentry_sdk.isolation_scope() as scope:
            with capture_internal_exceptions():
                scope.clear_breadcrumbs()
                scope.add_event_processor(
                    _make_request_event_processor(
                        gcp_event, configured_time, initial_time
                    )
                )
                scope.set_tag("gcp_region", environ.get("FUNCTION_REGION"))
                timeout_thread = None
                if (
                    integration.timeout_warning
                    and configured_time > TIMEOUT_WARNING_BUFFER
                ):
                    waiting_time = configured_time - TIMEOUT_WARNING_BUFFER

                    timeout_thread = TimeoutThread(
                        waiting_time,
                        configured_time,
                        isolation_scope=scope,
                        current_scope=sentry_sdk.get_current_scope(),
                    )

                    # Starting the thread to raise timeout warning exception
                    timeout_thread.start()

            headers = {}
            header_attributes: "dict[str, Any]" = {}
            if hasattr(gcp_event, "headers"):
                headers = gcp_event.headers
                for header, header_value in _filter_headers(
                    headers, use_annotated_value=False
                ).items():
                    header_attributes[f"http.request.header.{header.lower()}"] = (
                        # header_value will always be a string because we set `use_annotated_value` to false above
                        header_value
                    )

            additional_attributes = {}
            if hasattr(gcp_event, "method"):
                additional_attributes["http.request.method"] = gcp_event.method

            if hasattr(gcp_event, "query_string"):
                query_string = gcp_event.query_string.decode("utf-8", errors="replace")
                if query_string:
                    if has_data_collection_enabled(client.options):
                        filtered_qs = _apply_data_collection_filtering_to_query_string(
                            query_string=query_string,
                            behaviour=client.options["data_collection"][
                                "url_query_params"
                            ],
                        )
                        if filtered_qs:
                            additional_attributes["url.query"] = filtered_qs
                    elif should_send_default_pii():
                        additional_attributes["url.query"] = query_string

            sampling_context = {
                "gcp_env": {
                    "function_name": environ.get("FUNCTION_NAME"),
                    "function_entry_point": environ.get("ENTRY_POINT"),
                    "function_identity": environ.get("FUNCTION_IDENTITY"),
                    "function_region": environ.get("FUNCTION_REGION"),
                    "function_project": environ.get("GCP_PROJECT"),
                },
                "gcp_event": gcp_event,
            }

            function_name = environ.get("FUNCTION_NAME", "<unknown GCP function>")

            if environ.get("GCP_PROJECT"):
                additional_attributes["gcp.project.id"] = environ.get("GCP_PROJECT")

            if environ.get("FUNCTION_IDENTITY"):
                additional_attributes["faas.identity"] = environ.get(
                    "FUNCTION_IDENTITY"
                )

            if environ.get("ENTRY_POINT"):
                additional_attributes["faas.entry_point"] = environ.get("ENTRY_POINT")

            if has_span_streaming_enabled(client.options):
                sentry_sdk.traces.continue_trace(headers)
                Scope.set_custom_sampling_context(sampling_context)
                span_ctx = sentry_sdk.traces.start_span(
                    name=function_name,
                    parent_span=None,
                    attributes={
                        "sentry.op": OP.FUNCTION_GCP,
                        "sentry.origin": GcpIntegration.origin,
                        "sentry.segment.name.source": SegmentNameSource.COMPONENT,
                        "cloud.provider": CLOUD_PROVIDER.GCP,
                        "faas.name": function_name,
                        **header_attributes,
                        **additional_attributes,
                    },
                )
            else:
                transaction = continue_trace(
                    headers,
                    op=OP.FUNCTION_GCP,
                    name=environ.get("FUNCTION_NAME", ""),
                    source=TransactionSource.COMPONENT,
                    origin=GcpIntegration.origin,
                )

                span_ctx = sentry_sdk.start_transaction(
                    transaction, custom_sampling_context=sampling_context
                )

            with span_ctx:
                try:
                    return func(functionhandler, gcp_event, *args, **kwargs)
                except Exception:
                    exc_info = sys.exc_info()
                    sentry_event, hint = event_from_exception(
                        exc_info,
                        client_options=client.options,
                        mechanism={"type": "gcp", "handled": False},
                    )
                    sentry_sdk.capture_event(sentry_event, hint=hint)
                    reraise(*exc_info)
                finally:
                    if timeout_thread:
                        timeout_thread.stop()
                    # Flush out the event queue
                    client.flush()

    return sentry_func  # type: ignore


class GcpIntegration(Integration):
    identifier = "gcp"
    origin = f"auto.function.{identifier}"

    def __init__(self, timeout_warning: bool = False) -> None:
        self.timeout_warning = timeout_warning

    @staticmethod
    def setup_once() -> None:
        import __main__ as gcp_functions

        if not hasattr(gcp_functions, "worker_v1"):
            logger.warning(
                "GcpIntegration currently supports only Python 3.7 runtime environment."
            )
            return

        worker1 = gcp_functions.worker_v1

        worker1.FunctionHandler.invoke_user_function = _wrap_func(
            worker1.FunctionHandler.invoke_user_function
        )


def _make_request_event_processor(
    gcp_event: "Any", configured_timeout: "Any", initial_time: "Any"
) -> "EventProcessor":
    def event_processor(event: "Event", hint: "Hint") -> "Optional[Event]":
        final_time = datetime.now(timezone.utc)
        time_diff = final_time - initial_time

        execution_duration_in_millis = time_diff / timedelta(milliseconds=1)

        extra = event.setdefault("extra", {})
        extra["google cloud functions"] = {
            "function_name": environ.get("FUNCTION_NAME"),
            "function_entry_point": environ.get("ENTRY_POINT"),
            "function_identity": environ.get("FUNCTION_IDENTITY"),
            "function_region": environ.get("FUNCTION_REGION"),
            "function_project": environ.get("GCP_PROJECT"),
            "execution_duration_in_millis": execution_duration_in_millis,
            "configured_timeout_in_seconds": configured_timeout,
        }

        extra["google cloud logs"] = {
            "url": _get_google_cloud_logs_url(final_time),
        }

        request = event.get("request", {})

        request["url"] = "gcp:///{}".format(environ.get("FUNCTION_NAME"))

        if hasattr(gcp_event, "method"):
            request["method"] = gcp_event.method

        if hasattr(gcp_event, "query_string"):
            query_string = gcp_event.query_string.decode("utf-8", errors="replace")
            client_options = sentry_sdk.get_client().options
            if has_data_collection_enabled(client_options):
                if query_string:
                    filtered_qs = _apply_data_collection_filtering_to_query_string(
                        query_string=query_string,
                        behaviour=client_options["data_collection"]["url_query_params"],
                    )
                    if filtered_qs:
                        request["query_string"] = filtered_qs
            else:
                request["query_string"] = query_string

        if hasattr(gcp_event, "headers"):
            request["headers"] = _filter_headers(gcp_event.headers)

        if should_send_default_pii():
            if hasattr(gcp_event, "data"):
                request["data"] = gcp_event.data
        else:
            if hasattr(gcp_event, "data"):
                # Unfortunately couldn't find a way to get structured body from GCP
                # event. Meaning every body is unstructured to us.
                request["data"] = AnnotatedValue.removed_because_raw_data()

        event["request"] = deepcopy(request)

        return event

    return event_processor


def _get_google_cloud_logs_url(final_time: "datetime") -> str:
    """
    Generates a Google Cloud Logs console URL based on the environment variables
    Arguments:
        final_time {datetime} -- Final time
    Returns:
        str -- Google Cloud Logs Console URL to logs.
    """
    hour_ago = final_time - timedelta(hours=1)
    formatstring = "%Y-%m-%dT%H:%M:%SZ"

    url = (
        "https://console.cloud.google.com/logs/viewer?project={project}&resource=cloud_function"
        "%2Ffunction_name%2F{function_name}%2Fregion%2F{region}&minLogLevel=0&expandAll=false"
        "&timestamp={timestamp_end}&customFacets=&limitCustomFacetWidth=true"
        "&dateRangeStart={timestamp_start}&dateRangeEnd={timestamp_end}"
        "&interval=PT1H&scrollTimestamp={timestamp_end}"
    ).format(
        project=environ.get("GCP_PROJECT"),
        function_name=environ.get("FUNCTION_NAME"),
        region=environ.get("FUNCTION_REGION"),
        timestamp_end=final_time.strftime(formatstring),
        timestamp_start=hour_ago.strftime(formatstring),
    )

    return url



import copy
from threading import Lock
from typing import TYPE_CHECKING, Any

import sentry_sdk
from sentry_sdk._lru_cache import LRUCache
from sentry_sdk.tracing import Span
from sentry_sdk.tracing_utils import has_span_streaming_enabled

if TYPE_CHECKING:
    from typing import TypedDict

    FlagData = TypedDict("FlagData", {"flag": str, "result": bool})


DEFAULT_FLAG_CAPACITY = 100


class FlagBuffer:
    def __init__(self, capacity: int) -> None:
        self.capacity = capacity
        self.lock = Lock()

        # Buffer is private. The name is mangled to discourage use. If you use this attribute
        # directly you're on your own!
        self.__buffer = LRUCache(capacity)

    def clear(self) -> None:
        self.__buffer = LRUCache(self.capacity)

    def __deepcopy__(self, memo: "dict[int, Any]") -> "FlagBuffer":
        with self.lock:
            buffer = FlagBuffer(self.capacity)
            buffer.__buffer = copy.deepcopy(self.__buffer, memo)
            return buffer

    def get(self) -> "list[FlagData]":
        with self.lock:
            return [
                {"flag": key, "result": value} for key, value in self.__buffer.get_all()
            ]

    def set(self, flag: str, result: bool) -> None:
        if isinstance(result, FlagBuffer):
            # If someone were to insert `self` into `self` this would create a circular dependency
            # on the lock. This is of course a deadlock. However, this is far outside the expected
            # usage of this class. We guard against it here for completeness and to document this
            # expected failure mode.
            raise ValueError(
                "FlagBuffer instances can not be inserted into the dictionary."
            )

        with self.lock:
            self.__buffer.set(flag, result)


def add_feature_flag(flag: str, result: bool) -> None:
    """
    Records a flag and its value to be sent on subsequent error events.
    We recommend you do this on flag evaluations. Flags are buffered per Sentry scope.
    """
    client = sentry_sdk.get_client()

    flags = sentry_sdk.get_isolation_scope().flags
    flags.set(flag, result)

    if has_span_streaming_enabled(client.options):
        span = sentry_sdk.traces.get_current_span()
        if span and isinstance(span, sentry_sdk.traces.StreamedSpan):
            span.set_attribute(f"flag.evaluation.{flag}", result)

    else:
        span = sentry_sdk.get_current_span()
        if span and isinstance(span, Span):
            span.set_flag(f"flag.evaluation.{flag}", result)



from typing import TYPE_CHECKING

import sentry_sdk
from sentry_sdk.consts import OP, SPANDATA
from sentry_sdk.tracing_utils import has_span_streaming_enabled

from ..consts import SPAN_ORIGIN

if TYPE_CHECKING:
    import agents


def handoff_span(
    context: "agents.RunContextWrapper", from_agent: "agents.Agent", to_agent_name: str
) -> None:
    span_streaming = has_span_streaming_enabled(sentry_sdk.get_client().options)
    if span_streaming:
        with sentry_sdk.traces.start_span(
            name=f"handoff from {from_agent.name} to {to_agent_name}",
            attributes={
                "sentry.op": OP.GEN_AI_HANDOFF,
                "sentry.origin": SPAN_ORIGIN,
                SPANDATA.GEN_AI_OPERATION_NAME: "handoff",
            },
        ) as span:
            # Add conversation ID from agent
            conv_id = getattr(from_agent, "_sentry_conversation_id", None)
            if conv_id:
                span.set_attribute(SPANDATA.GEN_AI_CONVERSATION_ID, conv_id)
    else:
        with sentry_sdk.start_span(
            op=OP.GEN_AI_HANDOFF,
            name=f"handoff from {from_agent.name} to {to_agent_name}",
            origin=SPAN_ORIGIN,
        ) as span:
            span.set_data(SPANDATA.GEN_AI_OPERATION_NAME, "handoff")

            # Add conversation ID from agent
            conv_id = getattr(from_agent, "_sentry_conversation_id", None)
            if conv_id:
                span.set_data(SPANDATA.GEN_AI_CONVERSATION_ID, conv_id)



import sys
from functools import partial
from typing import TYPE_CHECKING

import sentry_sdk
from sentry_sdk._werkzeug import _get_headers, get_host
from sentry_sdk.api import continue_trace
from sentry_sdk.consts import OP, SPANDATA
from sentry_sdk.data_collection import _apply_data_collection_filtering_to_query_string
from sentry_sdk.integrations._wsgi_common import (
    DEFAULT_HTTP_METHODS_TO_CAPTURE,
    _filter_headers,
)
from sentry_sdk.scope import Scope, should_send_default_pii, use_isolation_scope
from sentry_sdk.sessions import track_session
from sentry_sdk.traces import SegmentNameSource, StreamedSpan
from sentry_sdk.tracing import Span, TransactionSource
from sentry_sdk.tracing_utils import has_span_streaming_enabled
from sentry_sdk.utils import (
    ContextVar,
    capture_internal_exceptions,
    event_from_exception,
    has_data_collection_enabled,
    nullcontext,
    reraise,
)

if TYPE_CHECKING:
    from typing import (
        Any,
        Callable,
        ContextManager,
        Dict,
        Iterator,
        Optional,
        Protocol,
        Tuple,
        TypeVar,
        Union,
    )

    from sentry_sdk._types import Event, EventProcessor
    from sentry_sdk.utils import ExcInfo

    WsgiResponseIter = TypeVar("WsgiResponseIter")
    WsgiResponseHeaders = TypeVar("WsgiResponseHeaders")
    WsgiExcInfo = TypeVar("WsgiExcInfo")

    class StartResponse(Protocol):
        def __call__(
            self,
            status: str,
            response_headers: "WsgiResponseHeaders",
            exc_info: "Optional[WsgiExcInfo]" = None,
        ) -> "WsgiResponseIter":  # type: ignore
            pass


_wsgi_middleware_applied = ContextVar("sentry_wsgi_middleware_applied")
_DEFAULT_TRANSACTION_NAME = "generic WSGI request"


def wsgi_decoding_dance(s: str, charset: str = "utf-8", errors: str = "replace") -> str:
    return s.encode("latin1").decode(charset, errors)


def get_request_url(
    environ: "Dict[str, str]", use_x_forwarded_for: bool = False
) -> str:
    """Return the absolute URL without query string for the given WSGI
    environment."""
    script_name = environ.get("SCRIPT_NAME", "").rstrip("/")
    path_info = environ.get("PATH_INFO", "").lstrip("/")
    path = f"{script_name}/{path_info}"

    scheme = environ.get("wsgi.url_scheme")
    if use_x_forwarded_for:
        scheme = environ.get("HTTP_X_FORWARDED_PROTO", scheme)

    return "%s://%s/%s" % (
        scheme,
        get_host(environ, use_x_forwarded_for),
        wsgi_decoding_dance(path).lstrip("/"),
    )


class SentryWsgiMiddleware:
    __slots__ = (
        "app",
        "use_x_forwarded_for",
        "span_origin",
        "http_methods_to_capture",
    )

    def __init__(
        self,
        app: "Callable[[Dict[str, str], Callable[..., Any]], Any]",
        use_x_forwarded_for: bool = False,
        span_origin: str = "manual",
        http_methods_to_capture: "Tuple[str, ...]" = DEFAULT_HTTP_METHODS_TO_CAPTURE,
    ) -> None:
        self.app = app
        self.use_x_forwarded_for = use_x_forwarded_for
        self.span_origin = span_origin
        self.http_methods_to_capture = http_methods_to_capture

    def __call__(
        self, environ: "Dict[str, str]", start_response: "Callable[..., Any]"
    ) -> "Any":
        if _wsgi_middleware_applied.get(False):
            return self.app(environ, start_response)

        client = sentry_sdk.get_client()
        span_streaming = has_span_streaming_enabled(client.options)

        _wsgi_middleware_applied.set(True)
        try:
            with sentry_sdk.isolation_scope() as scope:
                with track_session(scope, session_mode="request"):
                    with capture_internal_exceptions():
                        scope.clear_breadcrumbs()
                        scope._name = "wsgi"
                        scope.add_event_processor(
                            _make_wsgi_event_processor(
                                environ, self.use_x_forwarded_for
                            )
                        )

                    method = environ.get("REQUEST_METHOD", "").upper()

                    span_ctx: "Optional[ContextManager[Union[Span, StreamedSpan, None]]]" = None
                    if method in self.http_methods_to_capture:
                        if span_streaming:
                            sentry_sdk.traces.continue_trace(
                                dict(_get_headers(environ))
                            )
                            Scope.set_custom_sampling_context({"wsgi_environ": environ})

                            if has_data_collection_enabled(client.options):
                                if client.options["data_collection"]["user_info"]:
                                    client_ip = get_client_ip(environ)
                                    if client_ip:
                                        scope.set_attribute(
                                            SPANDATA.USER_IP_ADDRESS, client_ip
                                        )
                            elif should_send_default_pii():
                                client_ip = get_client_ip(environ)
                                if client_ip:
                                    scope.set_attribute(
                                        SPANDATA.USER_IP_ADDRESS, client_ip
                                    )

                            span_ctx = sentry_sdk.traces.start_span(
                                name=_DEFAULT_TRANSACTION_NAME,
                                attributes={
                                    "sentry.segment.name.source": SegmentNameSource.ROUTE,
                                    "sentry.origin": self.span_origin,
                                    "sentry.op": OP.HTTP_SERVER,
                                },
                                parent_span=None,
                            )
                        else:
                            transaction = continue_trace(
                                environ,
                                op=OP.HTTP_SERVER,
                                name=_DEFAULT_TRANSACTION_NAME,
                                source=TransactionSource.ROUTE,
                                origin=self.span_origin,
                            )

                            span_ctx = sentry_sdk.start_transaction(
                                transaction,
                                custom_sampling_context={"wsgi_environ": environ},
                            )

                    span_ctx = span_ctx or nullcontext()

                    with span_ctx as span:
                        if isinstance(span, StreamedSpan):
                            with capture_internal_exceptions():
                                for attr, value in _get_request_attributes(
                                    environ, self.use_x_forwarded_for
                                ).items():
                                    span.set_attribute(attr, value)

                        try:
                            response = self.app(
                                environ,
                                partial(_sentry_start_response, start_response, span),
                            )
                        except BaseException:
                            reraise(*_capture_exception())
        finally:
            _wsgi_middleware_applied.set(False)

        # Within the uWSGI subhandler, the use of the "offload" mechanism for file responses
        # is determined by a pointer equality check on the response object
        # (see https://github.com/unbit/uwsgi/blob/8d116f7ea2b098c11ce54d0b3a561c54dcd11929/plugins/python/wsgi_subhandler.c#L278).
        #
        # If we were to return a _ScopedResponse, this would cause the check to always fail
        # since it's checking the files are exactly the same.
        #
        # To avoid this and ensure that the offloading mechanism works as expected when it's
        # enabled, we check if the response is a file-like object (determined by the presence
        # of `fileno`), if the wsgi.file_wrapper is available in the environment (as if so,
        # it would've been used in handling the file in the response).
        #
        # Even if the offload mechanism is not enabled, there are optimizations that uWSGI does for file-like objects,
        # so we want to make sure we don't interfere with those either.
        #
        # If all conditions are met, we return the original response object directly,
        # allowing uWSGI to handle it as intended.
        if (
            environ.get("wsgi.file_wrapper")
            and getattr(response, "fileno", None) is not None
        ):
            return response

        return _ScopedResponse(scope, response)


def _sentry_start_response(
    old_start_response: "StartResponse",
    span: "Optional[Union[Span, StreamedSpan]]",
    status: str,
    response_headers: "WsgiResponseHeaders",
    exc_info: "Optional[WsgiExcInfo]" = None,
) -> "WsgiResponseIter":  # type: ignore[type-var]
    with capture_internal_exceptions():
        status_int = int(status.split(" ", 1)[0])
        if span is not None:
            if isinstance(span, StreamedSpan):
                span.status = "error" if status_int >= 400 else "ok"
                span.set_attribute("http.response.status_code", status_int)
            else:
                span.set_http_status(status_int)

    if exc_info is None:
        # The Django Rest Framework WSGI test client, and likely other
        # (incorrect) implementations, cannot deal with the exc_info argument
        # if one is present. Avoid providing a third argument if not necessary.
        return old_start_response(status, response_headers)
    else:
        return old_start_response(status, response_headers, exc_info)


def _get_environ(environ: "Dict[str, str]") -> "Iterator[Tuple[str, str]]":
    """
    Returns our explicitly included environment variables we want to
    capture (server name, port and remote addr if pii is enabled).
    """
    keys = ["SERVER_NAME", "SERVER_PORT"]
    client_options = sentry_sdk.get_client().options

    # make debugging of proxy setup easier. Proxy headers are
    # in headers.
    if has_data_collection_enabled(client_options):
        if client_options["data_collection"]["user_info"]:
            keys += ["REMOTE_ADDR"]
    elif should_send_default_pii():
        keys += ["REMOTE_ADDR"]

    for key in keys:
        if key in environ:
            yield key, environ[key]


def get_client_ip(environ: "Dict[str, str]") -> "Optional[Any]":
    """
    Infer the user IP address from various headers. This cannot be used in
    security sensitive situations since the value may be forged from a client,
    but it's good enough for the event payload.
    """
    try:
        return environ["HTTP_X_FORWARDED_FOR"].split(",")[0].strip()
    except (KeyError, IndexError):
        pass

    try:
        return environ["HTTP_X_REAL_IP"]
    except KeyError:
        pass

    return environ.get("REMOTE_ADDR")


def _capture_exception() -> "ExcInfo":
    """
    Captures the current exception and sends it to Sentry.
    Returns the ExcInfo tuple to it can be reraised afterwards.
    """
    exc_info = sys.exc_info()
    e = exc_info[1]

    # SystemExit(0) is the only uncaught exception that is expected behavior
    should_skip_capture = isinstance(e, SystemExit) and e.code in (0, None)
    if not should_skip_capture:
        event, hint = event_from_exception(
            exc_info,
            client_options=sentry_sdk.get_client().options,
            mechanism={"type": "wsgi", "handled": False},
        )
        sentry_sdk.capture_event(event, hint=hint)

    return exc_info


class _ScopedResponse:
    """
    Users a separate scope for each response chunk.

    This will make WSGI apps more tolerant against:
    - WSGI servers streaming responses from a different thread/from
      different threads than the one that called start_response
    - close() not being called
    - WSGI servers streaming responses interleaved from the same thread
    """

    __slots__ = ("_response", "_scope")

    def __init__(
        self, scope: "sentry_sdk.scope.Scope", response: "Iterator[bytes]"
    ) -> None:
        self._scope = scope
        self._response = response

    def __iter__(self) -> "Iterator[bytes]":
        iterator = iter(self._response)

        while True:
            with use_isolation_scope(self._scope):
                try:
                    chunk = next(iterator)
                except StopIteration:
                    break
                except BaseException:
                    reraise(*_capture_exception())

            yield chunk

    def close(self) -> None:
        with use_isolation_scope(self._scope):
            try:
                self._response.close()  # type: ignore
            except AttributeError:
                pass
            except BaseException:
                reraise(*_capture_exception())


def _make_wsgi_event_processor(
    environ: "Dict[str, str]", use_x_forwarded_for: bool
) -> "EventProcessor":
    # It's a bit unfortunate that we have to extract and parse the request data
    # from the environ so eagerly, but there are a few good reasons for this.
    #
    # We might be in a situation where the scope never gets torn down
    # properly. In that case we will have an unnecessary strong reference to
    # all objects in the environ (some of which may take a lot of memory) when
    # we're really just interested in a few of them.
    #
    # Keeping the environment around for longer than the request lifecycle is
    # also not necessarily something uWSGI can deal with:
    # https://github.com/unbit/uwsgi/issues/1950

    client_ip = get_client_ip(environ)
    request_url = get_request_url(environ, use_x_forwarded_for)
    query_string = environ.get("QUERY_STRING")
    method = environ.get("REQUEST_METHOD")
    env = dict(_get_environ(environ))
    headers = _filter_headers(dict(_get_headers(environ)))
    client_options = sentry_sdk.get_client().options

    def event_processor(event: "Event", hint: "Dict[str, Any]") -> "Event":
        with capture_internal_exceptions():
            # if the code below fails halfway through we at least have some data
            request_info = event.setdefault("request", {})

            if has_data_collection_enabled(client_options):
                if client_options["data_collection"]["user_info"]:
                    user_info = event.setdefault("user", {})
                    if client_ip:
                        user_info.setdefault("ip_address", client_ip)
            elif should_send_default_pii():
                user_info = event.setdefault("user", {})
                if client_ip:
                    user_info.setdefault("ip_address", client_ip)

            request_info["url"] = request_url
            request_info["method"] = method
            request_info["env"] = env
            request_info["headers"] = headers

            if has_data_collection_enabled(client_options):
                if query_string:
                    filtered_qs = _apply_data_collection_filtering_to_query_string(
                        query_string=query_string,
                        behaviour=client_options["data_collection"]["url_query_params"],
                    )
                    if filtered_qs:
                        request_info["query_string"] = filtered_qs
            else:
                # This was not originally gated so if data collection is not enabled, leave as-is.
                request_info["query_string"] = query_string

        return event

    return event_processor


def _get_request_attributes(
    environ: "Dict[str, str]",
    use_x_forwarded_for: bool = False,
) -> "Dict[str, Any]":
    """
    Return span attributes related to the HTTP request from the WSGI environ.
    """
    attributes: "dict[str, Any]" = {}

    method = environ.get("REQUEST_METHOD")
    if method:
        attributes["http.request.method"] = method.upper()

    headers = _filter_headers(dict(_get_headers(environ)), use_annotated_value=False)
    for header, value in headers.items():
        attributes[f"http.request.header.{header.lower()}"] = value

    url_scheme = environ.get("wsgi.url_scheme")
    if url_scheme:
        attributes["network.protocol.name"] = url_scheme

    server_name = environ.get("SERVER_NAME")
    if server_name:
        attributes["server.address"] = server_name

    server_port = environ.get("SERVER_PORT")
    if server_port:
        try:
            attributes["server.port"] = int(server_port)
        except ValueError:
            pass

    client_options = sentry_sdk.get_client().options

    if has_data_collection_enabled(client_options):
        query_string = environ.get("QUERY_STRING")
        filtered_qs = None
        if query_string:
            filtered_qs = _apply_data_collection_filtering_to_query_string(
                query_string=query_string,
                behaviour=client_options["data_collection"]["url_query_params"],
            )

            if filtered_qs:
                attributes["http.query"] = filtered_qs

        path = environ.get("PATH_INFO", "")
        if path:
            attributes["url.path"] = path

        attributes["url.full"] = get_request_url(environ, use_x_forwarded_for)
        if filtered_qs is not None:
            attributes["url.full"] += f"?{filtered_qs}"

        if client_options["data_collection"]["user_info"]:
            client_ip = get_client_ip(environ)
            if client_ip:
                attributes["client.address"] = client_ip

    elif should_send_default_pii():
        client_ip = get_client_ip(environ)
        if client_ip:
            attributes["client.address"] = client_ip

        query_string = environ.get("QUERY_STRING")
        if query_string:
            attributes["http.query"] = query_string

        path = environ.get("PATH_INFO", "")
        if path:
            attributes["url.path"] = path

        url_full = get_request_url(environ, use_x_forwarded_for)
        if query_string:
            url_full += "?" + query_string
        attributes["url.full"] = url_full

    return attributes



import functools
import os
import sys
import weakref

import sentry_sdk
from sentry_sdk.integrations import DidNotEnable, Integration
from sentry_sdk.integrations._wsgi_common import RequestExtractor
from sentry_sdk.integrations.wsgi import SentryWsgiMiddleware
from sentry_sdk.scope import should_send_default_pii
from sentry_sdk.traces import SOURCE_FOR_STYLE as SEGMENT_SOURCE_FOR_STYLE
from sentry_sdk.tracing import SOURCE_FOR_STYLE as TRANSACTION_SOURCE_FOR_STYLE
from sentry_sdk.tracing_utils import has_span_streaming_enabled
from sentry_sdk.utils import (
    capture_internal_exceptions,
    ensure_integration_enabled,
    event_from_exception,
    has_data_collection_enabled,
    reraise,
)

try:
    from pyramid.httpexceptions import HTTPException
    from pyramid.request import Request
except ImportError:
    raise DidNotEnable("Pyramid not installed")

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Callable, Dict, Optional

    from pyramid.response import Response
    from webob.cookies import RequestCookies
    from webob.request import _FieldStorageWithFile

    from sentry_sdk._types import Event, EventProcessor
    from sentry_sdk.integrations.wsgi import _ScopedResponse
    from sentry_sdk.utils import ExcInfo


if getattr(Request, "authenticated_userid", None):

    def authenticated_userid(request: "Request") -> "Optional[Any]":
        return request.authenticated_userid

else:
    # bw-compat for pyramid < 1.5
    from pyramid.security import authenticated_userid  # type: ignore


TRANSACTION_STYLE_VALUES = ("route_name", "route_pattern")


class PyramidIntegration(Integration):
    identifier = "pyramid"
    origin = f"auto.http.{identifier}"

    transaction_style = ""

    def __init__(self, transaction_style: str = "route_name") -> None:
        if transaction_style not in TRANSACTION_STYLE_VALUES:
            raise ValueError(
                "Invalid value for transaction_style: %s (must be in %s)"
                % (transaction_style, TRANSACTION_STYLE_VALUES)
            )
        self.transaction_style = transaction_style

    @staticmethod
    def setup_once() -> None:
        from pyramid import router

        old_call_view = router._call_view

        @functools.wraps(old_call_view)
        def sentry_patched_call_view(
            registry: "Any", request: "Request", *args: "Any", **kwargs: "Any"
        ) -> "Response":
            client = sentry_sdk.get_client()
            integration = client.get_integration(PyramidIntegration)
            if integration is None:
                return old_call_view(registry, request, *args, **kwargs)

            _set_transaction_name_and_source(
                sentry_sdk.get_current_scope(), integration.transaction_style, request
            )

            scope = sentry_sdk.get_isolation_scope()

            if has_span_streaming_enabled(client.options):
                if has_data_collection_enabled(client.options):
                    if client.options["data_collection"]["user_info"]:
                        user_id = authenticated_userid(request)
                        if user_id:
                            scope.set_user({"id": user_id})
                elif should_send_default_pii():
                    user_id = authenticated_userid(request)
                    if user_id:
                        scope.set_user({"id": user_id})

            scope.add_event_processor(
                _make_event_processor(weakref.ref(request), integration)
            )

            return old_call_view(registry, request, *args, **kwargs)

        router._call_view = sentry_patched_call_view

        if hasattr(Request, "invoke_exception_view"):
            old_invoke_exception_view = Request.invoke_exception_view

            def sentry_patched_invoke_exception_view(
                self: "Request", *args: "Any", **kwargs: "Any"
            ) -> "Any":
                rv = old_invoke_exception_view(self, *args, **kwargs)

                if (
                    self.exc_info
                    and all(self.exc_info)
                    and rv.status_int == 500
                    and sentry_sdk.get_client().get_integration(PyramidIntegration)
                    is not None
                ):
                    _capture_exception(self.exc_info)

                return rv

            Request.invoke_exception_view = sentry_patched_invoke_exception_view

        old_wsgi_call = router.Router.__call__

        @ensure_integration_enabled(PyramidIntegration, old_wsgi_call)
        def sentry_patched_wsgi_call(
            self: "Any", environ: "Dict[str, str]", start_response: "Callable[..., Any]"
        ) -> "_ScopedResponse":
            def sentry_patched_inner_wsgi_call(
                environ: "Dict[str, Any]", start_response: "Callable[..., Any]"
            ) -> "Any":
                try:
                    return old_wsgi_call(self, environ, start_response)
                except Exception:
                    einfo = sys.exc_info()
                    _capture_exception(einfo)
                    reraise(*einfo)

            middleware = SentryWsgiMiddleware(
                sentry_patched_inner_wsgi_call,
                span_origin=PyramidIntegration.origin,
            )
            return middleware(environ, start_response)

        router.Router.__call__ = sentry_patched_wsgi_call


@ensure_integration_enabled(PyramidIntegration)
def _capture_exception(exc_info: "ExcInfo") -> None:
    if exc_info[0] is None or issubclass(exc_info[0], HTTPException):
        return

    event, hint = event_from_exception(
        exc_info,
        client_options=sentry_sdk.get_client().options,
        mechanism={"type": "pyramid", "handled": False},
    )

    sentry_sdk.capture_event(event, hint=hint)


def _set_transaction_name_and_source(
    scope: "sentry_sdk.Scope", transaction_style: str, request: "Request"
) -> None:
    try:
        name_for_style = {
            "route_name": request.matched_route.name,
            "route_pattern": request.matched_route.pattern,
        }
        is_span_streaming_enabled = has_span_streaming_enabled(
            sentry_sdk.get_client().options
        )
        source = (
            SEGMENT_SOURCE_FOR_STYLE[transaction_style]
            if is_span_streaming_enabled
            else TRANSACTION_SOURCE_FOR_STYLE[transaction_style]
        )
        scope.set_transaction_name(
            name_for_style[transaction_style],
            source=source,
        )
    except Exception:
        pass


class PyramidRequestExtractor(RequestExtractor):
    def url(self) -> str:
        return self.request.path_url

    def env(self) -> "Dict[str, str]":
        return self.request.environ

    def cookies(self) -> "RequestCookies":
        return self.request.cookies

    def raw_data(self) -> str:
        return self.request.text

    def form(self) -> "Dict[str, str]":
        return {
            key: value
            for key, value in self.request.POST.items()
            if not getattr(value, "filename", None)
        }

    def files(self) -> "Dict[str, _FieldStorageWithFile]":
        return {
            key: value
            for key, value in self.request.POST.items()
            if getattr(value, "filename", None)
        }

    def size_of_file(self, postdata: "_FieldStorageWithFile") -> int:
        file = postdata.file
        try:
            return os.fstat(file.fileno()).st_size
        except Exception:
            return 0


def _make_event_processor(
    weak_request: "Callable[[], Request]", integration: "PyramidIntegration"
) -> "EventProcessor":
    def pyramid_event_processor(event: "Event", hint: "Dict[str, Any]") -> "Event":
        request = weak_request()
        if request is None:
            return event

        with capture_internal_exceptions():
            PyramidRequestExtractor(request).extract_into_event(event)

        client_options = sentry_sdk.get_client().options
        if has_data_collection_enabled(client_options):
            if client_options["data_collection"]["user_info"]:
                with capture_internal_exceptions():
                    user_info = event.setdefault("user", {})
                    user_info.setdefault("id", authenticated_userid(request))
        elif should_send_default_pii():
            with capture_internal_exceptions():
                user_info = event.setdefault("user", {})
                user_info.setdefault("id", authenticated_userid(request))

        return event

    return pyramid_event_processor



import sys
from typing import TYPE_CHECKING

from sentry_sdk.consts import SPANDATA
from sentry_sdk.integrations import DidNotEnable
from sentry_sdk.traces import StreamedSpan
from sentry_sdk.utils import capture_internal_exceptions, reraise

from ..spans import (
    handoff_span,
    invoke_agent_span,
    update_invoke_agent_span,
)

if TYPE_CHECKING:
    from typing import Any, Awaitable, Callable, Optional, Union

    from agents.run_internal.run_steps import SingleStepResult

    from sentry_sdk.tracing import Span

try:
    import agents
except ImportError:
    raise DidNotEnable("OpenAI Agents not installed")


def _has_active_agent_span(context_wrapper: "agents.RunContextWrapper") -> bool:
    """Check if there's an active agent span for this context"""
    return getattr(context_wrapper, "_sentry_current_agent", None) is not None


def _get_current_agent(
    context_wrapper: "agents.RunContextWrapper",
) -> "Optional[agents.Agent]":
    """Get the current agent from context wrapper"""
    return getattr(context_wrapper, "_sentry_current_agent", None)


def _close_streaming_workflow_span(agent: "Optional[agents.Agent]") -> None:
    """Close the workflow span for streaming executions if it exists."""
    if agent and hasattr(agent, "_sentry_workflow_span"):
        workflow_span = agent._sentry_workflow_span
        workflow_span.__exit__(*sys.exc_info())
        delattr(agent, "_sentry_workflow_span")


def _maybe_start_agent_span(
    context_wrapper: "agents.RunContextWrapper",
    agent: "agents.Agent",
    should_run_agent_start_hooks: bool,
    span_kwargs: "dict[str, Any]",
    is_streaming: bool = False,
) -> "Optional[Union[Span, StreamedSpan]]":
    """
    Start an agent invocation span if conditions are met.
    Handles ending any existing span for a different agent.

    Returns the new span if started, or the existing span if conditions aren't met.
    """
    if not (should_run_agent_start_hooks and agent and context_wrapper):
        return getattr(context_wrapper, "_sentry_agent_span", None)

    # End any existing span for a different agent
    if _has_active_agent_span(context_wrapper):
        current_agent = _get_current_agent(context_wrapper)
        if current_agent and current_agent != agent:
            span = getattr(context_wrapper, "_sentry_agent_span", None)
            if span:
                update_invoke_agent_span(
                    span=span, context=context_wrapper, agent=agent
                )
                span.__exit__(None, None, None)
                delattr(context_wrapper, "_sentry_agent_span")

    # Store the agent on the context wrapper so we can access it later
    context_wrapper._sentry_current_agent = agent
    span = invoke_agent_span(context_wrapper, agent, span_kwargs)
    context_wrapper._sentry_agent_span = span
    agent._sentry_agent_span = span

    if not is_streaming:
        return span

    if isinstance(span, StreamedSpan):
        span.set_attribute(SPANDATA.GEN_AI_RESPONSE_STREAMING, True)
    else:
        span.set_data(SPANDATA.GEN_AI_RESPONSE_STREAMING, True)

    return span


async def _run_single_turn(
    original_run_single_turn: "Callable[..., Awaitable[SingleStepResult]]",
    *args: "Any",
    **kwargs: "Any",
) -> "SingleStepResult":
    """
    Patched _run_single_turn that
    - creates agent invocation spans if there is no already active agent invocation span.
    - ends the agent invocation span if and only if an exception is raised in `_run_single_turn()`.
    """
    # openai-agents >= 0.14 passes `bindings: AgentBindings` instead of `agent`.
    bindings = kwargs.get("bindings")
    agent = (
        getattr(bindings, "public_agent", None)
        if bindings is not None
        else kwargs.get("agent")
    )
    context_wrapper = kwargs.get("context_wrapper")
    should_run_agent_start_hooks = kwargs.get("should_run_agent_start_hooks", False)

    span = _maybe_start_agent_span(
        context_wrapper, agent, should_run_agent_start_hooks, kwargs
    )

    if (
        span is None
        or (isinstance(span, StreamedSpan) and span.end_timestamp is not None)
        or (not isinstance(span, StreamedSpan) and span.timestamp is not None)
    ):
        return await original_run_single_turn(*args, **kwargs)

    try:
        result = await original_run_single_turn(*args, **kwargs)
    except Exception:
        exc_info = sys.exc_info()
        with capture_internal_exceptions():
            span = getattr(context_wrapper, "_sentry_agent_span", None)
            if span:
                update_invoke_agent_span(
                    span=span, context=context_wrapper, agent=agent
                )
                span.__exit__(*exc_info)
                delattr(context_wrapper, "_sentry_agent_span")
        reraise(*exc_info)

    return result


async def _run_single_turn_streamed(
    original_run_single_turn_streamed: "Callable[..., Awaitable[SingleStepResult]]",
    *args: "Any",
    **kwargs: "Any",
) -> "SingleStepResult":
    """
    Patched _run_single_turn_streamed that
    - creates agent invocation spans for streaming if there is no already active agent invocation span.
    - ends the agent invocation span if and only if `_run_single_turn_streamed()` raises an exception.

    Note: Unlike _run_single_turn which uses keyword-only arguments (*,),
    _run_single_turn_streamed uses positional arguments. The call signature <v0.14 is:
    _run_single_turn_streamed(
        streamed_result,              # args[0]
        agent,                        # args[1]
        hooks,                        # args[2]
        context_wrapper,              # args[3]
        run_config,                   # args[4]
        should_run_agent_start_hooks, # args[5]
        tool_use_tracker,             # args[6]
        all_tools,                    # args[7]
        server_conversation_tracker,  # args[8] (optional)
    )

    The call signature >=v0.14 is:
    _run_single_turn_streamed(
        streamed_result,              # args[0]
        bindings,                     # args[1]
        hooks,                        # args[2]
        context_wrapper,              # args[3]
        run_config,                   # args[4]
        should_run_agent_start_hooks, # args[5]
        tool_use_tracker,             # args[6]
        all_tools,                    # args[7]
        server_conversation_tracker,  # args[8] (optional)
    )
    """
    streamed_result = args[0] if len(args) > 0 else kwargs.get("streamed_result")
    # openai-agents >= 0.14 passes `bindings: AgentBindings` at args[1] instead of `agent`.
    agent_or_bindings = (
        args[1] if len(args) > 1 else kwargs.get("bindings", kwargs.get("agent"))
    )
    agent = getattr(agent_or_bindings, "public_agent", agent_or_bindings)
    context_wrapper = args[3] if len(args) > 3 else kwargs.get("context_wrapper")
    should_run_agent_start_hooks = bool(
        args[5] if len(args) > 5 else kwargs.get("should_run_agent_start_hooks", False)
    )

    span_kwargs: "dict[str, Any]" = {}
    if streamed_result and hasattr(streamed_result, "input"):
        span_kwargs["original_input"] = streamed_result.input

    span = _maybe_start_agent_span(
        context_wrapper,
        agent,
        should_run_agent_start_hooks,
        span_kwargs,
        is_streaming=True,
    )

    if (
        span is None
        or (isinstance(span, StreamedSpan) and span.end_timestamp is not None)
        or (not isinstance(span, StreamedSpan) and span.timestamp is not None)
    ):
        return await original_run_single_turn_streamed(*args, **kwargs)

    try:
        result = await original_run_single_turn_streamed(*args, **kwargs)
    except Exception:
        exc_info = sys.exc_info()
        with capture_internal_exceptions():
            span = getattr(context_wrapper, "_sentry_agent_span", None)
            if span:
                update_invoke_agent_span(
                    span=span, context=context_wrapper, agent=agent
                )
                span.__exit__(*exc_info)
                delattr(context_wrapper, "_sentry_agent_span")
            _close_streaming_workflow_span(agent)
        reraise(*exc_info)

    return result


async def _execute_handoffs(
    original_execute_handoffs: "Callable[..., SingleStepResult]",
    *args: "Any",
    **kwargs: "Any",
) -> "SingleStepResult":
    """
    Patched execute_handoffs that
    - creates and manages handoff spans.
    - ends the agent invocation span.
    - ends the workflow span if the response is streamed and an exception is raised in `execute_handoffs()`.
    """

    context_wrapper = kwargs.get("context_wrapper")
    run_handoffs = kwargs.get("run_handoffs")
    # openai-agents >= 0.14 renamed `agent` to `public_agent`.
    agent = kwargs.get("public_agent", kwargs.get("agent"))

    # Create Sentry handoff span for the first handoff (agents library only processes the first one)
    if run_handoffs:
        first_handoff = run_handoffs[0]
        handoff_agent_name = first_handoff.handoff.agent_name
        handoff_span(context_wrapper, agent, handoff_agent_name)

    if not agent or not context_wrapper or not _has_active_agent_span(context_wrapper):
        # Call original method with all parameters
        try:
            return await original_execute_handoffs(*args, **kwargs)
        except Exception:
            exc_info = sys.exc_info()
            with capture_internal_exceptions():
                _close_streaming_workflow_span(agent)
            reraise(*exc_info)

    # Call original method with all parameters
    try:
        result = await original_execute_handoffs(*args, **kwargs)
    except Exception:
        exc_info = sys.exc_info()
        with capture_internal_exceptions():
            _close_streaming_workflow_span(agent)
            span = getattr(context_wrapper, "_sentry_agent_span", None)
            if span:
                update_invoke_agent_span(
                    span=span, context=context_wrapper, agent=agent
                )
                span.__exit__(*exc_info)
                delattr(context_wrapper, "_sentry_agent_span")
        reraise(*exc_info)

    span = getattr(context_wrapper, "_sentry_agent_span", None)
    if span:
        update_invoke_agent_span(span=span, context=context_wrapper, agent=agent)
        span.__exit__(None, None, None)
        delattr(context_wrapper, "_sentry_agent_span")

    return result


async def _execute_final_output(
    original_execute_final_output: "Callable[..., SingleStepResult]",
    *args: "Any",
    **kwargs: "Any",
) -> "SingleStepResult":
    """
    Patched execute_final_output that
    - ends the agent invocation span.
    - ends the workflow span if the response is streamed.
    """

    # openai-agents >= 0.14 renamed `agent` to `public_agent`.
    agent = kwargs.get("public_agent", kwargs.get("agent"))
    context_wrapper = kwargs.get("context_wrapper")
    final_output = kwargs.get("final_output")

    if not agent or not context_wrapper or not _has_active_agent_span(context_wrapper):
        try:
            return await original_execute_final_output(*args, **kwargs)
        finally:
            with capture_internal_exceptions():
                # For streaming, close the workflow span (non-streaming uses context manager in _create_run_wrapper)
                _close_streaming_workflow_span(agent)

    try:
        result = await original_execute_final_output(*args, **kwargs)
    except Exception:
        exc_info = sys.exc_info()
        with capture_internal_exceptions():
            # For streaming, close the workflow span (non-streaming uses context manager in _create_run_wrapper)
            _close_streaming_workflow_span(agent)
            span = getattr(context_wrapper, "_sentry_agent_span", None)
            if span:
                update_invoke_agent_span(
                    span=span, context=context_wrapper, agent=agent, output=final_output
                )
                span.__exit__(*exc_info)
                delattr(context_wrapper, "_sentry_agent_span")
        reraise(*exc_info)

    span = getattr(context_wrapper, "_sentry_agent_span", None)
    if span:
        update_invoke_agent_span(
            span=span, context=context_wrapper, agent=agent, output=final_output
        )
        span.__exit__(None, None, None)
        delattr(context_wrapper, "_sentry_agent_span")

    return result



import functools
from typing import TYPE_CHECKING

from django import VERSION as DJANGO_VERSION
from django.core.cache import CacheHandler
from urllib3.util import parse_url as urlparse

import sentry_sdk
from sentry_sdk.consts import OP, SPANDATA
from sentry_sdk.integrations.redis.utils import _get_safe_key, _key_as_string
from sentry_sdk.tracing_utils import has_span_streaming_enabled
from sentry_sdk.utils import (
    capture_internal_exceptions,
    ensure_integration_enabled,
)

if TYPE_CHECKING:
    from typing import Any, Callable, Optional


METHODS_TO_INSTRUMENT = [
    "set",
    "set_many",
    "get",
    "get_many",
]


def _get_span_description(
    method_name: str, args: "tuple[Any]", kwargs: "dict[str, Any]"
) -> str:
    return _key_as_string(_get_safe_key(method_name, args, kwargs))


def _patch_cache_method(
    cache: "CacheHandler",
    method_name: str,
    address: "Optional[str]",
    port: "Optional[int]",
) -> None:
    from sentry_sdk.integrations.django import DjangoIntegration

    original_method = getattr(cache, method_name)

    @ensure_integration_enabled(DjangoIntegration, original_method)
    def _instrument_call(
        cache: "CacheHandler",
        method_name: str,
        original_method: "Callable[..., Any]",
        args: "tuple[Any, ...]",
        kwargs: "dict[str, Any]",
        address: "Optional[str]",
        port: "Optional[int]",
    ) -> "Any":
        is_set_operation = method_name.startswith("set")
        is_get_method = method_name == "get"
        is_get_many_method = method_name == "get_many"

        op = OP.CACHE_PUT if is_set_operation else OP.CACHE_GET
        description = _get_span_description(method_name, args, kwargs)

        span_streaming = has_span_streaming_enabled(sentry_sdk.get_client().options)
        if span_streaming:
            if sentry_sdk.traces.get_current_span() is None:
                return original_method(*args, **kwargs)
            with sentry_sdk.traces.start_span(
                name=description,
                attributes={
                    "sentry.op": op,
                    "sentry.origin": DjangoIntegration.origin,
                },
            ) as span:
                value = original_method(*args, **kwargs)

                with capture_internal_exceptions():
                    if address is not None:
                        span.set_attribute(SPANDATA.NETWORK_PEER_ADDRESS, address)

                    if port is not None:
                        span.set_attribute(SPANDATA.NETWORK_PEER_PORT, port)

                    key = _get_safe_key(method_name, args, kwargs)
                    if key is not None:
                        span.set_attribute(SPANDATA.CACHE_KEY, key)

                    item_size = None
                    if is_get_many_method:
                        if value != {}:
                            item_size = len(str(value))
                            span.set_attribute(SPANDATA.CACHE_HIT, True)
                        else:
                            span.set_attribute(SPANDATA.CACHE_HIT, False)
                    elif is_get_method:
                        default_value = None
                        if len(args) >= 2:
                            default_value = args[1]
                        elif "default" in kwargs:
                            default_value = kwargs["default"]

                        if value != default_value:
                            item_size = len(str(value))
                            span.set_attribute(SPANDATA.CACHE_HIT, True)
                        else:
                            span.set_attribute(SPANDATA.CACHE_HIT, False)
                    else:  # TODO: We don't handle `get_or_set` which we should
                        arg_count = len(args)
                        if arg_count >= 2:
                            # 'set' command
                            item_size = len(str(args[1]))
                        elif arg_count == 1:
                            # 'set_many' command
                            item_size = len(str(args[0]))

                    if item_size is not None:
                        span.set_attribute(SPANDATA.CACHE_ITEM_SIZE, item_size)

                return value
        else:
            with sentry_sdk.start_span(
                op=op,
                name=description,
                origin=DjangoIntegration.origin,
            ) as span:
                value = original_method(*args, **kwargs)

                with capture_internal_exceptions():
                    if address is not None:
                        span.set_data(SPANDATA.NETWORK_PEER_ADDRESS, address)

                    if port is not None:
                        span.set_data(SPANDATA.NETWORK_PEER_PORT, port)

                    key = _get_safe_key(method_name, args, kwargs)
                    if key is not None:
                        span.set_data(SPANDATA.CACHE_KEY, key)

                    item_size = None
                    if is_get_many_method:
                        if value != {}:
                            item_size = len(str(value))
                            span.set_data(SPANDATA.CACHE_HIT, True)
                        else:
                            span.set_data(SPANDATA.CACHE_HIT, False)
                    elif is_get_method:
                        default_value = None
                        if len(args) >= 2:
                            default_value = args[1]
                        elif "default" in kwargs:
                            default_value = kwargs["default"]

                        if value != default_value:
                            item_size = len(str(value))
                            span.set_data(SPANDATA.CACHE_HIT, True)
                        else:
                            span.set_data(SPANDATA.CACHE_HIT, False)
                    else:  # TODO: We don't handle `get_or_set` which we should
                        arg_count = len(args)
                        if arg_count >= 2:
                            # 'set' command
                            item_size = len(str(args[1]))
                        elif arg_count == 1:
                            # 'set_many' command
                            item_size = len(str(args[0]))

                    if item_size is not None:
                        span.set_data(SPANDATA.CACHE_ITEM_SIZE, item_size)

                return value

    @functools.wraps(original_method)
    def sentry_method(*args: "Any", **kwargs: "Any") -> "Any":
        return _instrument_call(
            cache, method_name, original_method, args, kwargs, address, port
        )

    setattr(cache, method_name, sentry_method)


def _patch_cache(
    cache: "CacheHandler", address: "Optional[str]" = None, port: "Optional[int]" = None
) -> None:
    if not hasattr(cache, "_sentry_patched"):
        for method_name in METHODS_TO_INSTRUMENT:
            _patch_cache_method(cache, method_name, address, port)
        cache._sentry_patched = True


def _get_address_port(
    settings: "dict[str, Any]",
) -> "tuple[Optional[str], Optional[int]]":
    location = settings.get("LOCATION")

    # TODO: location can also be an array of locations
    #       see: https://docs.djangoproject.com/en/5.0/topics/cache/#redis
    #       GitHub issue: https://github.com/getsentry/sentry-python/issues/3062
    if not isinstance(location, str):
        return None, None

    if "://" in location:
        parsed_url = urlparse(location)
        # remove the username and password from URL to not leak sensitive data.
        address = "{}://{}{}".format(
            parsed_url.scheme or "",
            parsed_url.hostname or "",
            parsed_url.path or "",
        )
        port = parsed_url.port
    else:
        address = location
        port = None

    return address, int(port) if port is not None else None


def should_enable_cache_spans() -> bool:
    from sentry_sdk.integrations.django import DjangoIntegration

    client = sentry_sdk.get_client()
    integration = client.get_integration(DjangoIntegration)
    from django.conf import settings

    return integration is not None and (
        (client.spotlight is not None and settings.DEBUG is True)
        or integration.cache_spans is True
    )


def patch_caching() -> None:
    if not hasattr(CacheHandler, "_sentry_patched"):
        if DJANGO_VERSION < (3, 2):
            original_get_item = CacheHandler.__getitem__

            @functools.wraps(original_get_item)
            def sentry_get_item(self: "CacheHandler", alias: str) -> "Any":
                cache = original_get_item(self, alias)

                if should_enable_cache_spans():
                    from django.conf import settings

                    address, port = _get_address_port(
                        settings.CACHES[alias or "default"]
                    )

                    _patch_cache(cache, address, port)

                return cache

            CacheHandler.__getitem__ = sentry_get_item
            CacheHandler._sentry_patched = True

        else:
            original_create_connection = CacheHandler.create_connection

            @functools.wraps(original_create_connection)
            def sentry_create_connection(self: "CacheHandler", alias: str) -> "Any":
                cache = original_create_connection(self, alias)

                if should_enable_cache_spans():
                    address, port = _get_address_port(self.settings[alias or "default"])

                    _patch_cache(cache, address, port)

                return cache

            CacheHandler.create_connection = sentry_create_connection
            CacheHandler._sentry_patched = True



from typing import TYPE_CHECKING

import sentry_sdk
from sentry_sdk.consts import OP, SPANDATA
from sentry_sdk.integrations.redis.consts import (
    SPAN_ORIGIN,
)
from sentry_sdk.integrations.redis.modules.caches import (
    _compile_cache_span_properties,
    _set_cache_data,
)
from sentry_sdk.integrations.redis.modules.queries import _compile_db_span_properties
from sentry_sdk.integrations.redis.utils import (
    _extract_key,
    _get_safe_command,
    _set_client_data,
    _set_pipeline_data,
)
from sentry_sdk.tracing import Span
from sentry_sdk.tracing_utils import has_span_streaming_enabled
from sentry_sdk.utils import capture_internal_exceptions

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Any, Optional, Union

    from sentry_sdk.traces import StreamedSpan


def patch_redis_pipeline(
    pipeline_cls: "Any",
    is_cluster: bool,
    get_command_args_fn: "Any",
    set_db_data_fn: "Callable[[Union[Span, StreamedSpan], Any], None]",
) -> None:
    old_execute = pipeline_cls.execute

    from sentry_sdk.integrations.redis import RedisIntegration

    def sentry_patched_execute(self: "Any", *args: "Any", **kwargs: "Any") -> "Any":
        client = sentry_sdk.get_client()
        if client.get_integration(RedisIntegration) is None:
            return old_execute(self, *args, **kwargs)

        sentry_sdk.add_breadcrumb(
            message="redis.pipeline.execute",
            type="redis",
            category="redis",
            data={
                "redis.is_cluster": is_cluster,
                "redis.transaction": False if is_cluster else self.transaction,
            },
        )

        span_streaming = has_span_streaming_enabled(client.options)
        span: "Union[Span, StreamedSpan]"
        if span_streaming:
            if sentry_sdk.traces.get_current_span() is None:
                return old_execute(self, *args, **kwargs)
            span = sentry_sdk.traces.start_span(
                name="redis.pipeline.execute",
                attributes={
                    "sentry.origin": SPAN_ORIGIN,
                    "sentry.op": OP.DB_REDIS,
                },
            )
        else:
            span = sentry_sdk.start_span(
                op=OP.DB_REDIS,
                name="redis.pipeline.execute",
                origin=SPAN_ORIGIN,
            )

        with span:
            with capture_internal_exceptions():
                command_seq = None
                try:
                    command_seq = self._execution_strategy.command_queue
                except AttributeError:
                    command_seq = self.command_stack

                set_db_data_fn(span, self)
                _set_pipeline_data(
                    span,
                    is_cluster,
                    get_command_args_fn,
                    False if is_cluster else self.transaction,
                    command_seq,
                )

            return old_execute(self, *args, **kwargs)

    pipeline_cls.execute = sentry_patched_execute


def patch_redis_client(
    cls: "Any",
    is_cluster: bool,
    set_db_data_fn: "Callable[[Union[Span, StreamedSpan], Any], None]",
) -> None:
    """
    This function can be used to instrument custom redis client classes or
    subclasses.
    """
    old_execute_command = cls.execute_command

    from sentry_sdk.integrations.redis import RedisIntegration

    def sentry_patched_execute_command(
        self: "Any", name: str, *args: "Any", **kwargs: "Any"
    ) -> "Any":
        client = sentry_sdk.get_client()
        integration = client.get_integration(RedisIntegration)
        if integration is None:
            return old_execute_command(self, name, *args, **kwargs)

        db_properties = _compile_db_span_properties(integration, name, args)

        breadcrumb_data = {
            "redis.is_cluster": is_cluster,
            "redis.command": name,
            "db.operation": name,
        }
        key = _extract_key(name, args)
        if key is not None:
            breadcrumb_data["redis.key"] = key

        sentry_sdk.add_breadcrumb(
            message=db_properties["description"],
            type="redis",
            category="redis",
            data=breadcrumb_data,
        )

        span_streaming = has_span_streaming_enabled(client.options)

        if span_streaming and sentry_sdk.traces.get_current_span() is None:
            return old_execute_command(self, name, *args, **kwargs)

        cache_properties = _compile_cache_span_properties(
            name,
            args,
            kwargs,
            integration,
        )

        additional_cache_span_attributes = {}
        with capture_internal_exceptions():
            additional_cache_span_attributes[SPANDATA.DB_QUERY_TEXT] = (
                _get_safe_command(name, args)
            )

        cache_span: "Optional[Union[Span, StreamedSpan]]" = None
        if cache_properties["is_cache_key"] and cache_properties["op"] is not None:
            if span_streaming:
                cache_span = sentry_sdk.traces.start_span(
                    name=cache_properties["description"],
                    attributes={
                        "sentry.op": cache_properties["op"],
                        "sentry.origin": SPAN_ORIGIN,
                        **additional_cache_span_attributes,
                    },
                )
            else:
                cache_span = sentry_sdk.start_span(
                    op=cache_properties["op"],
                    name=cache_properties["description"],
                    origin=SPAN_ORIGIN,
                )
            cache_span.__enter__()

        additional_db_span_attributes = {}
        with capture_internal_exceptions():
            additional_db_span_attributes[SPANDATA.DB_QUERY_TEXT] = _get_safe_command(
                name, args
            )

        db_span: "Union[Span, StreamedSpan]"
        if span_streaming:
            db_span = sentry_sdk.traces.start_span(
                name=db_properties["description"],
                attributes={
                    "sentry.op": db_properties["op"],
                    "sentry.origin": SPAN_ORIGIN,
                    **additional_db_span_attributes,
                },
            )
        else:
            db_span = sentry_sdk.start_span(
                op=db_properties["op"],
                name=db_properties["description"],
                origin=SPAN_ORIGIN,
            )
        db_span.__enter__()

        set_db_data_fn(db_span, self)
        _set_client_data(db_span, is_cluster, name, *args)

        value = old_execute_command(self, name, *args, **kwargs)

        db_span.__exit__(None, None, None)

        if cache_span:
            _set_cache_data(cache_span, self, cache_properties, value)
            cache_span.__exit__(None, None, None)

        return value

    cls.execute_command = sentry_patched_execute_command



from typing import TYPE_CHECKING, Any

from sentry_sdk.feature_flags import add_feature_flag
from sentry_sdk.integrations import DidNotEnable, Integration

try:
    from openfeature import api
    from openfeature.hook import Hook

    if TYPE_CHECKING:
        from openfeature.hook import HookContext, HookHints
except ImportError:
    raise DidNotEnable("OpenFeature is not installed")


class OpenFeatureIntegration(Integration):
    identifier = "openfeature"

    @staticmethod
    def setup_once() -> None:
        # Register the hook within the global openfeature hooks list.
        api.add_hooks(hooks=[OpenFeatureHook()])


class OpenFeatureHook(Hook):
    def after(self, hook_context: "Any", details: "Any", hints: "Any") -> None:
        if isinstance(details.value, bool):
            add_feature_flag(details.flag_key, details.value)

    def error(
        self, hook_context: "HookContext", exception: Exception, hints: "HookHints"
    ) -> None:
        if isinstance(hook_context.default_value, bool):
            add_feature_flag(hook_context.flag_key, hook_context.default_value)



"""
This integration ingests tracing data from native extensions written in Rust.

Using it requires additional setup on the Rust side to accept a
`RustTracingLayer` Python object and register it with the `tracing-subscriber`
using an adapter from the `pyo3-python-tracing-subscriber` crate. For example:
```rust
#[pyfunction]
pub fn initialize_tracing(py_impl: Bound<'_, PyAny>) {
    tracing_subscriber::registry()
        .with(pyo3_python_tracing_subscriber::PythonCallbackLayerBridge::new(py_impl))
        .init();
}
```

Usage in Python would then look like:
```
sentry_sdk.init(
    dsn=sentry_dsn,
    integrations=[
        RustTracingIntegration(
            "demo_rust_extension",
            demo_rust_extension.initialize_tracing,
            event_type_mapping=event_type_mapping,
        )
    ],
)
```

Each native extension requires its own integration.
"""

import json
from enum import Enum, auto
from typing import Any, Callable, Dict, Optional, Union

import sentry_sdk
from sentry_sdk.integrations import Integration
from sentry_sdk.scope import should_send_default_pii
from sentry_sdk.traces import StreamedSpan
from sentry_sdk.tracing import Span as SentrySpan
from sentry_sdk.tracing_utils import has_span_streaming_enabled
from sentry_sdk.utils import SENSITIVE_DATA_SUBSTITUTE


class RustTracingLevel(Enum):
    Trace = "TRACE"
    Debug = "DEBUG"
    Info = "INFO"
    Warn = "WARN"
    Error = "ERROR"


class EventTypeMapping(Enum):
    Ignore = auto()
    Exc = auto()
    Breadcrumb = auto()
    Event = auto()


def tracing_level_to_sentry_level(level: str) -> "sentry_sdk._types.LogLevelStr":
    level = RustTracingLevel(level)
    if level in (RustTracingLevel.Trace, RustTracingLevel.Debug):
        return "debug"
    elif level == RustTracingLevel.Info:
        return "info"
    elif level == RustTracingLevel.Warn:
        return "warning"
    elif level == RustTracingLevel.Error:
        return "error"
    else:
        # Better this than crashing
        return "info"


def extract_contexts(event: "Dict[str, Any]") -> "Dict[str, Any]":
    metadata = event.get("metadata", {})
    contexts = {}

    location = {}
    for field in ["module_path", "file", "line"]:
        if field in metadata:
            location[field] = metadata[field]
    if len(location) > 0:
        contexts["rust_tracing_location"] = location

    fields = {}
    for field in metadata.get("fields", []):
        fields[field] = event.get(field)
    if len(fields) > 0:
        contexts["rust_tracing_fields"] = fields

    return contexts


def process_event(event: "Dict[str, Any]") -> None:
    metadata = event.get("metadata", {})

    logger = metadata.get("target")
    level = tracing_level_to_sentry_level(metadata.get("level"))
    message: "sentry_sdk._types.Any" = event.get("message")
    contexts = extract_contexts(event)

    sentry_event: "sentry_sdk._types.Event" = {
        "logger": logger,
        "level": level,
        "message": message,
        "contexts": contexts,
    }

    sentry_sdk.capture_event(sentry_event)


def process_exception(event: "Dict[str, Any]") -> None:
    process_event(event)


def process_breadcrumb(event: "Dict[str, Any]") -> None:
    level = tracing_level_to_sentry_level(event.get("metadata", {}).get("level"))
    message = event.get("message")

    sentry_sdk.add_breadcrumb(level=level, message=message)


def default_span_filter(metadata: "Dict[str, Any]") -> bool:
    return RustTracingLevel(metadata.get("level")) in (
        RustTracingLevel.Error,
        RustTracingLevel.Warn,
        RustTracingLevel.Info,
    )


def default_event_type_mapping(metadata: "Dict[str, Any]") -> "EventTypeMapping":
    level = RustTracingLevel(metadata.get("level"))
    if level == RustTracingLevel.Error:
        return EventTypeMapping.Exc
    elif level in (RustTracingLevel.Warn, RustTracingLevel.Info):
        return EventTypeMapping.Breadcrumb
    elif level in (RustTracingLevel.Debug, RustTracingLevel.Trace):
        return EventTypeMapping.Ignore
    else:
        return EventTypeMapping.Ignore


class RustTracingLayer:
    def __init__(
        self,
        origin: str,
        event_type_mapping: """Callable[
            [Dict[str, Any]], EventTypeMapping
        ]""" = default_event_type_mapping,
        span_filter: "Callable[[Dict[str, Any]], bool]" = default_span_filter,
        include_tracing_fields: "Optional[bool]" = None,
    ):
        self.origin = origin
        self.event_type_mapping = event_type_mapping
        self.span_filter = span_filter
        self.include_tracing_fields = include_tracing_fields

    def _include_tracing_fields(self) -> bool:
        """
        By default, the values of tracing fields are not included in case they
        contain PII. A user may override that by passing `True` for the
        `include_tracing_fields` keyword argument of this integration or by
        setting `send_default_pii` to `True` in their Sentry client options.
        """
        return (
            should_send_default_pii()
            if self.include_tracing_fields is None
            else self.include_tracing_fields
        )

    def on_event(self, event: str, sentry_span: "SentrySpan") -> None:
        deserialized_event = json.loads(event)
        metadata = deserialized_event.get("metadata", {})

        event_type = self.event_type_mapping(metadata)
        if event_type == EventTypeMapping.Ignore:
            return
        elif event_type == EventTypeMapping.Exc:
            process_exception(deserialized_event)
        elif event_type == EventTypeMapping.Breadcrumb:
            process_breadcrumb(deserialized_event)
        elif event_type == EventTypeMapping.Event:
            process_event(deserialized_event)

    def on_new_span(
        self, attrs: str, span_id: str
    ) -> "Optional[Union[SentrySpan, StreamedSpan]]":
        attrs = json.loads(attrs)
        metadata = attrs.get("metadata", {})

        if not self.span_filter(metadata):
            return None

        module_path = metadata.get("module_path")
        name = metadata.get("name")
        message = attrs.get("message")

        if message is not None:
            sentry_span_name = message
        elif module_path is not None and name is not None:
            sentry_span_name = f"{module_path}::{name}"  # noqa: E231
        elif name is not None:
            sentry_span_name = name
        else:
            sentry_span_name = "<unknown>"

        span_streaming = has_span_streaming_enabled(sentry_sdk.get_client().options)
        if span_streaming:
            if sentry_sdk.traces.get_current_span() is None:
                return None

            sentry_span = sentry_sdk.traces.start_span(
                name=sentry_span_name,
                attributes={
                    "sentry.op": "function",
                    "sentry.origin": self.origin,
                },
            )
            fields = metadata.get("fields", [])
            for field in fields:
                if self._include_tracing_fields():
                    sentry_span.set_attribute(field, attrs.get(field))
                else:
                    sentry_span.set_attribute(field, SENSITIVE_DATA_SUBSTITUTE)

            return sentry_span

        sentry_span = sentry_sdk.start_span(
            op="function",
            name=sentry_span_name,
            origin=self.origin,
        )
        fields = metadata.get("fields", [])
        for field in fields:
            if self._include_tracing_fields():
                sentry_span.set_data(field, attrs.get(field))
            else:
                sentry_span.set_data(field, SENSITIVE_DATA_SUBSTITUTE)

        sentry_span.__enter__()
        return sentry_span

    def on_close(self, span_id: str, sentry_span: "SentrySpan") -> None:
        if sentry_span is None:
            return

        sentry_span.__exit__(None, None, None)

    def on_record(
        self, span_id: str, values: str, sentry_span: "Union[SentrySpan, StreamedSpan]"
    ) -> None:
        if sentry_span is None:
            return

        set_on_span = (
            sentry_span.set_attribute
            if isinstance(sentry_span, StreamedSpan)
            else sentry_span.set_data
        )

        deserialized_values = json.loads(values)
        for key, value in deserialized_values.items():
            if self._include_tracing_fields():
                set_on_span(key, value)
            else:
                set_on_span(key, SENSITIVE_DATA_SUBSTITUTE)


class RustTracingIntegration(Integration):
    """
    Ingests tracing data from a Rust native extension's `tracing` instrumentation.

    If a project uses more than one Rust native extension, each one will need
    its own instance of `RustTracingIntegration` with an initializer function
    specific to that extension.

    Since all of the setup for this integration requires instance-specific state
    which is not available in `setup_once()`, setup instead happens in `__init__()`.
    """

    def __init__(
        self,
        identifier: str,
        initializer: "Callable[[RustTracingLayer], None]",
        event_type_mapping: """Callable[
            [Dict[str, Any]], EventTypeMapping
        ]""" = default_event_type_mapping,
        span_filter: "Callable[[Dict[str, Any]], bool]" = default_span_filter,
        include_tracing_fields: "Optional[bool]" = None,
    ):
        self.identifier = identifier
        origin = f"auto.function.rust_tracing.{identifier}"
        self.tracing_layer = RustTracingLayer(
            origin, event_type_mapping, span_filter, include_tracing_fields
        )

        initializer(self.tracing_layer)

    @staticmethod
    def setup_once() -> None:
        pass



import functools

from sentry_sdk.integrations import DidNotEnable, Integration
from sentry_sdk.utils import capture_internal_exceptions, parse_version

try:
    import pydantic_ai  # noqa: F401
    from pydantic_ai import Agent
except ImportError:
    raise DidNotEnable("pydantic-ai not installed")


from importlib.metadata import PackageNotFoundError, version
from typing import TYPE_CHECKING

from .patches import (
    _patch_agent_run,
    _patch_graph_nodes,
    _patch_tool_execution,
)
from .spans.ai_client import ai_client_span, update_ai_client_span

if TYPE_CHECKING:
    from typing import Any

    from pydantic_ai import ModelRequestContext, RunContext
    from pydantic_ai.capabilities import Hooks
    from pydantic_ai.messages import ModelResponse


def register_hooks(hooks: "Hooks") -> None:
    """
    Creates hooks for chat model calls and register the hooks by adding the hooks to the `capabilities` argument passed to `Agent.__init__()`.
    """

    @hooks.on.before_model_request
    async def on_request(
        ctx: "RunContext[None]", request_context: "ModelRequestContext"
    ) -> "ModelRequestContext":
        run_context_metadata = ctx.metadata
        if not isinstance(run_context_metadata, dict):
            return request_context

        span = ai_client_span(
            messages=request_context.messages,
            agent=None,
            model=request_context.model,
            model_settings=request_context.model_settings,
        )

        run_context_metadata["_sentry_span"] = span
        span.__enter__()

        return request_context

    @hooks.on.after_model_request
    async def on_response(
        ctx: "RunContext[None]",
        *,
        request_context: "ModelRequestContext",
        response: "ModelResponse",
    ) -> "ModelResponse":
        run_context_metadata = ctx.metadata
        if not isinstance(run_context_metadata, dict):
            return response

        span = run_context_metadata.pop("_sentry_span", None)
        if span is None:
            return response

        update_ai_client_span(span, response)
        span.__exit__(None, None, None)

        return response

    @hooks.on.model_request_error
    async def on_error(
        ctx: "RunContext[None]",
        *,
        request_context: "ModelRequestContext",
        error: "Exception",
    ) -> "ModelResponse":
        run_context_metadata = ctx.metadata

        if not isinstance(run_context_metadata, dict):
            raise error

        span = run_context_metadata.pop("_sentry_span", None)
        if span is None:
            raise error

        with capture_internal_exceptions():
            span.__exit__(type(error), error, error.__traceback__)

        raise error

    original_init = Agent.__init__

    @functools.wraps(original_init)
    def patched_init(self: "Agent[Any, Any]", *args: "Any", **kwargs: "Any") -> None:
        caps = list(kwargs.get("capabilities") or [])
        caps.append(hooks)
        kwargs["capabilities"] = caps

        metadata = kwargs.get("metadata")
        if metadata is None:
            kwargs["metadata"] = {}  # Used as shared reference between hooks

        return original_init(self, *args, **kwargs)

    Agent.__init__ = patched_init  # type: ignore[method-assign]


class PydanticAIIntegration(Integration):
    """
    Typical interaction with the library:
    1. The user creates an Agent instance with configuration, including system instructions sent to every model call.
    2. The user calls `Agent.run()` or `Agent.run_stream()` to start an agent run. The latter can be used to incrementally receive progress.
    - Each run invocation has `RunContext` objects that are passed to the library hooks.
    3. In a loop, the agent repeatedly calls the model, maintaining a conversation history that includes previous messages and tool results, which is passed to each call.

    Internally, Pydantic AI maintains an execution graph in which ModelRequestNode are responsible for model calls, including retries.
    Hooks using the decorators provided by `pydantic_ai.capabilities` create and manage spans for model calls when these hooks are available (newer library versions).
    The span is created in `on_request` and stored in the metadata of the `RunContext` object shared with `on_response` and `on_error`.

    The metadata dictionary on the RunContext instance is initialized with `{"_sentry_span": None}` in the `_create_run_wrapper()` and `_create_streaming_wrapper()` wrappers that
    instrument `Agent.run()` and `Agent.run_stream()`, respectively. A non-empty dictionary is required for the metadata object to be a shared reference between hooks.
    """

    identifier = "pydantic_ai"
    origin = f"auto.ai.{identifier}"
    using_request_hooks = False

    def __init__(
        self, include_prompts: bool = True, handled_tool_call_exceptions: bool = True
    ) -> None:
        """
        Initialize the Pydantic AI integration.

        Args:
            include_prompts: Whether to include prompts and messages in span data.
                Requires send_default_pii=True. Defaults to True.
            handled_tool_exceptions: Capture tool call exceptions that Pydantic AI
                internally prevents from bubbling up.
        """
        self.include_prompts = include_prompts
        self.handled_tool_call_exceptions = handled_tool_call_exceptions

    @staticmethod
    def setup_once() -> None:
        """
        Set up the pydantic-ai integration.

        This patches the key methods in pydantic-ai to create Sentry spans for:
        - Agent invocations (Agent.run methods)
        - Model requests (AI client calls)
        - Tool executions
        """
        _patch_agent_run()
        _patch_tool_execution()

        PydanticAIIntegration.using_request_hooks = False
        try:
            PYDANTIC_AI_VERSION = version("pydantic-ai-slim")
        except PackageNotFoundError:
            return

        PYDANTIC_AI_VERSION = parse_version(PYDANTIC_AI_VERSION)
        if PYDANTIC_AI_VERSION is None:
            return

        # ModelRequestContext.model added in https://github.com/pydantic/pydantic-ai/commit/f1260dfe09907f17688eee1646daf898fc428d4c
        if PYDANTIC_AI_VERSION < (
            1,
            73,
        ):
            _patch_graph_nodes()
            return

        try:
            from pydantic_ai.capabilities import Hooks
        except ImportError:
            return

        PydanticAIIntegration.using_request_hooks = True
        hooks = Hooks()
        register_hooks(hooks)
